# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from collections import defaultdict
from functools import lru_cache
from random import choice
from typing import Union, List

import numpy as np
from quafu import QuantumCircuit
from quafu.elements import Barrier, XYResonance, Measure
from quafu.elements.element_gates import SwapGate

from qsteed.dag.circuit_dag_convert import circuit_to_dag, gate_to_node, copy_dag
from qsteed.dag.dagcircuit import DAGCircuit
from qsteed.graph.couplinggraph import CouplingGraph
from qsteed.passes.basepass import BasePass
from qsteed.passes.datadict import DataDict
from qsteed.passes.mapping.baselayout import Layout


class SabreRouting(BasePass):
    """Execute the SABER routing algorithm once.

    References:
        Gushu Li, Yufei Ding, and Yuan Xie. "Tackling the qubit mapping problem
        for NISQ-era quantum devices." ASPLOS ’19, 1001–1014 (ACM, 2019).
        URL https://doi.org/10.1145/3297858.3304023
    """

    def __init__(self,
                 coupling_list: List = None,
                 heuristic="distance",
                 initial_layout: Layout = None,
                 modify_dag=False,
                 decay_delta: float = 0.01,
                 decay_reset_interval: int = 5,
                 extended_set_size: int = 20,
                 extended_set_weight: float = 0.5,
                 ):
        """SabreRouting initializer.

        Args:
            coupling_list (list): qubit coupling structure, e.g. [(0,1,0.99), (1,0,0.99), (1,2,0.98),(2,1,0.95)]
            heuristic (str): 'distance': The original SABER algorithm is based on the swap gates distance;
                             'fidelity': Noise-aware SABER algorithm based on two-qubit gates fidelity;
                             'mixture': First based on the 'distance', when the distance cannot be determined,
                                        based on the two-qubit gate 'fidelity'.
            initial_layout (Layout): The initial layout of the SABER routing algorithm.
            modify_dag (bool): The default is False, that is, the SWAP gate is not actually added to the
                                original dag, and the routing algorithm only changes the layout.
            decay_delta (float): The decay factor for penalizing serial swaps.
            decay_reset_interval (int): The number of swaps to apply before resetting the decay factor.
            extended_set_size (int): The size of the extended set.
            extended_set_weight (float): Weight of extended set compared to front_layer.
        """

        super().__init__()

        self.coupling_list = coupling_list
        self.coupling_graph = None
        self.initial_layout = initial_layout
        self.heuristic = heuristic
        self.modify_dag = modify_dag
        self.distance_matrix = None
        self.path_fidelity = None
        self.model = None
        self.add_swap_count = 0

        self.decay_delta = decay_delta
        self.decay_reset_interval = decay_reset_interval
        self.extended_set_size = extended_set_size
        self.extended_set_weight = extended_set_weight
        self.qubits_decay = None
        self.pre_executed_counts = None

        self.nodes_label = []

    def set_model(self, model):
        """Set the model, including information such as backend and layout.

        Args:
            model (Model): The given model includes information such as backend and layout.
        """
        self.model = model
        self.coupling_graph = self.model.get_backend().get_property('coupling_graph')

        if self.coupling_graph is None:
            if self.coupling_list is not None:
                coupling_graph = CouplingGraph(self.coupling_list)
                if coupling_graph.is_bidirectional is False:
                    coupling_graph.do_bidirectional()
                self.coupling_graph = coupling_graph
            else:
                raise ValueError("Error: There is no qubits coupling structure.")

        if self.model.datadict is None:
            self.model.datadict = DataDict()

    def get_model(self):
        """Get the model, including information such as backend and layout.
        """
        return self.model

    def run(self, circ_dag: Union[QuantumCircuit, DAGCircuit]):
        """Run the SabreRouting pass on QuantumCircuit or DAGCircuit.

        Args:
            circ_dag (DAGCircuit or QuantumCircuit): a dag or circuit to perform routing algorithm
        Returns:
            DAGCircuit: The original dag or the mapped_dag with added swap gate.
        """
        if isinstance(circ_dag, QuantumCircuit):
            dag = circuit_to_dag(circ_dag)
        elif isinstance(circ_dag, DAGCircuit):
            # dag = copy.deepcopy(circ_dag)
            dag = copy_dag(circ_dag)
        else:
            raise TypeError('Error: SabreSwap pass only supports QuantumCircuit or DAGCircuit.')

        if len(dag.qubits_used) == 1:  # Single-qubit circuit do not need to perform sabre
            return dag

        if float('inf') in dag.nodes:
            dag.remove_node(float('inf'))

        if len(dag.qubits_used) > self.coupling_graph.num_qubits:
            raise ValueError("More virtual qubits than physical qubits.")

        # In order to select SWAPs that can be executed in parallel,
        # a decay effect is introduced in the heuristic cost function.
        # If a qubit q_i is involved in a SWAP recently,
        # then its decay parameter will increase by δ, that is, decay(q_i) = 1 + δ.
        # TODO: dag.qubits_used is the actual qubit used. If the original circuit has an idle qubit line,
        #  this qubit will be lost, but it needs to be retained.
        self.qubits_decay = {qubit: 1 for qubit in dag.qubits_used}

        # Size of lookahead window. Set to number of qubits
        self.extended_set_size = self.coupling_graph.num_qubits

        # Decay coefficient for penalizing serial swaps. Set to average fidelity
        if self.heuristic == 'fidelity':
            self.decay_delta = 1 - sum(item for item in self.coupling_graph.edge_dict.values()) / len(
                self.coupling_graph.edge_dict)

        # How often to reset all decay rates to 1.
        self.decay_reset_interval = round(self.extended_set_size / 2)

        self.add_swap_count = 0  # Add number of SWAP gates.

        self.distance_matrix = self.coupling_graph.distance_matrix
        self.path_fidelity = self.coupling_graph.path_fidelity

        # Initialize mapped_dag, retaining only the name and qubit number information of the input original DAG.
        mapped_dag = None
        if self.modify_dag:
            mapped_dag = DAGCircuit()
            mapped_dag.name = dag.name
            mapped_dag.update_circuit_qubits(dag.circuit_qubits)

        if self.initial_layout is not None:
            self.model.set_layout({'initial_layout': self.initial_layout})
        elif self.model.get_layout()["initial_layout"] is None:
            layout = Layout()
            # Method1: Choose a trivial initial_layout.
            # layout.generate_trivial_layout(virtual_qubits=dag.circuit_qubits)
            # Method2: Choose a random initial_layout.
            layout.generate_random_layout(len(dag.qubits_used), self.coupling_graph.num_qubits)
            self.model.set_layout({'initial_layout': layout})

        # Copy initial_layout. Initialize Layout first and then copy 'initial_layout' through
        # set_layout() function, which is faster than deepcopy.
        # current_layout = copy.deepcopy(self.model.get_layout()["initial_layout"])
        current_layout = Layout()
        current_layout.set_layout(self.model.get_layout()["initial_layout"])

        # Parameter preparation before iteration.
        self.pre_executed_counts = defaultdict(int)  # executed counts
        front_layer = []
        for edge in dag.out_edges():
            if edge[0] == -1:
                self.pre_executed_counts[edge[1]] += 1
                if edge[1].name not in (Barrier.name, Measure.name):
                    if self.pre_executed_counts[edge[1]] == 2 or len(edge[1].pos) == 1:
                        front_layer.append(edge[1])
            else:
                break

        executed_2gate_list = []  # The hardware execution order list of executable 2-qubit gates under current_layout.
        unavailable_2qubits = set()

        # Start the algorithm from the front layer and iterate until all gates are completed.
        iteration_count = 0
        while front_layer:
            execute_gate_list = []

            # Delete executable gate
            for node in front_layer:
                if len(node.pos) == 2 and node.name not in [Barrier.name, XYResonance.name, Measure.name]:
                    v0, v1 = node.pos
                    p0, p1 = current_layout.v2p[v0], current_layout.v2p[v1]
                    if self.coupling_graph.graph.has_edge(p0, p1):
                        execute_gate_list.append(node)
                        # Record the coupling 2-qubits that have been used in the circuit.
                        unavailable_2qubits_new = {item for item in unavailable_2qubits if
                                                   p0 not in item and p1 not in item}
                        unavailable_2qubits = unavailable_2qubits_new
                else:  # Single-qubit gates, barriers, XY-gates and measures are both executable gates.
                    execute_gate_list.append(node)

            if execute_gate_list:
                for node in execute_gate_list:
                    self._apply_gate(mapped_dag, node, current_layout)
                    front_layer.remove(node)
                    for successor in self._dag_successors(dag, node):
                        if isinstance(successor.pos, dict):
                            self.pre_executed_counts[successor] += len(node.pos)
                            # self.pre_executed_counts[successor] += 1
                        else:
                            self.pre_executed_counts[successor] += 1
                        if self.pre_executed_counts[successor] == len(successor.pos):
                            # If all predecessors of a node in dag are applied, add this node to the front_layer.
                            front_layer.append(successor)

                # Reset qubits decay factors when a logical gate is applied.
                iteration_count = 0
                self._reset_qubits_decay()

                continue

            # When all available gates are exhausted, heuristically find the best swap and insert it.
            extended_set = self._calc_extended_set(dag, front_layer)
            swap_candidates = self._obtain_swaps(front_layer, current_layout)
            best_swap = self._get_best_swap(swap_candidates, current_layout, front_layer, extended_set,
                                            unavailable_2qubits)
            swap_node = gate_to_node(SwapGate(best_swap[0], best_swap[1]),
                                     specific_label='add' + str(self.add_swap_count))
            self._apply_gate(mapped_dag, swap_node, current_layout)
            self.add_swap_count += 1
            current_layout.swap(best_swap[0], best_swap[1])

            # Update the unavailable_2qubits set to prevent the same qubits from being inserted
            # into swap continuously, because two consecutive swaps on the same qubits can be eliminated.
            executed_2gate_list.append(
                (tuple(sorted((current_layout.v2p[best_swap[0]], current_layout.v2p[best_swap[1]]))), 1))
            unavailable_2qubits.add(
                tuple(sorted((current_layout.v2p[best_swap[0]], current_layout.v2p[best_swap[1]]))))

            # Update qubits_decay
            iteration_count += 1
            if iteration_count % self.decay_reset_interval == 0:
                self._reset_qubits_decay()
            else:
                self.qubits_decay[best_swap[0]] += self.decay_delta
                self.qubits_decay[best_swap[1]] += self.decay_delta

        # Update model data
        self.model.get_layout()["final_layout"] = current_layout
        self.model.datadict['add_swap_count'] = self.add_swap_count

        if self.modify_dag:
            return mapped_dag
        return dag

    def _apply_gate(self, mapped_dag, node, current_layout):
        """Add a node (gate) to dag.

        Args:
            mapped_dag (DAGCircuit): a real-time updated dag.
            node (dag.instruction_node.InstructionNode): The gate that needs to be inserted into the mapped_dag.
            current_layout (Layout): current layout
        """
        if not self.modify_dag:
            return

        # # Remove duplicate nodes
        # if node.label in self.nodes_label:
        #     return
        # else:
        #     self.nodes_label.append(node.label)

        mapped_op_node = copy.copy(node)
        if isinstance(node.pos, dict):
            mapped_qargs = {}
            for qubit, cbit in node.pos.items():
                mapped_qargs[current_layout.v2p[qubit]] = cbit
        else:
            mapped_qargs = [current_layout.v2p[pos] for pos in node.pos]
        mapped_op_node.pos = mapped_qargs
        mapped_dag.add_instruction_node_end(mapped_op_node)

    def _reset_qubits_decay(self):
        """Reset all qubit decay factors to 1.
        """
        self.qubits_decay = {k: 1 for k in self.qubits_decay.keys()}

    @lru_cache(maxsize=128)
    def _dag_successors(self, dag, node):
        """Find the successors of a node in the dag.

        Args:
            dag (DAGCircuit): a dag
            node: A node in the dag.
        Returns:
            successors: All successors of a node
        """
        successors = []
        for _, successor in dag.out_edges(node):
            successors.append(successor)
        return successors

    def _calc_extended_set(self, dag, front_layer):
        """Calculate the extended set for lookahead capabilities.

        Args:
            dag (DAGCircuit): a dag
            front_layer (list): The front layer in the dag.
        Returns:
            extended_set (set): Set of expansion gates obtained according to requirements.
        """
        extended_set = set()
        new_front_layer = list(front_layer)
        while len(new_front_layer) > 0 and len(extended_set) < self.extended_set_size:
            node = new_front_layer.pop(0)
            successors_nodes = self._dag_successors(dag, node)
            successors_nodes = [sn for sn in successors_nodes if sn.name not in [Barrier.name, Measure.name]]
            new_front_layer.extend(successors_nodes)
            new_front_layer = list(set(new_front_layer))
            for successor in successors_nodes:
                if len(successor.pos) == 2 and isinstance(successor.pos, list):
                    extended_set.update([successor])
        return extended_set

    def _obtain_swaps(self, front_layer, current_layout):
        """Get the set of candidate swap gates.

        Args:
            front_layer (list): front_layer gates list
            current_layout (Layout): current layout
        Returns:
            candidate_swaps (set): Set of candidate swap gates
        """
        candidate_swaps = set()
        for node in front_layer:
            for virtual in node.pos:
                physical = current_layout.v2p[virtual]
                for neighbor in self.coupling_graph.neighbors(physical):
                    virtual_neighbor = current_layout.p2v[neighbor]
                    swap = sorted([virtual, virtual_neighbor])
                    candidate_swaps.add(tuple(swap))

        return candidate_swaps

    def _swap_score(self, physical_swap_qubits):
        """Obtain the cost of the swap gate for the heuristic for 'fidelity'.

        Args:
            physical_swap_qubits (tuple): The physical qubits of the swap gate.
        Returns:
            swap_cost (float): The swap_cost calculated based on the coupled two-qubit fidelity.
        """
        min_f = min(np.log(self.coupling_graph.edge_dict[(physical_swap_qubits[0], physical_swap_qubits[1])]), np.log(
            self.coupling_graph.edge_dict[(physical_swap_qubits[1], physical_swap_qubits[0])]))
        max_f = max(np.log(self.coupling_graph.edge_dict[(physical_swap_qubits[0], physical_swap_qubits[1])]), np.log(
            self.coupling_graph.edge_dict[(physical_swap_qubits[1], physical_swap_qubits[0])]))
        swap_cost = 2 * max_f + min_f
        return swap_cost

    def _get_best_swap(self, swap_candidates, current_layout, front_layer, extended_set, unavailable_2qubits):
        """Get the best swap based on different heuristics.

        Args:
            swap_candidates (set): The set of all candidate swap gates.
            current_layout (Layout): current layout
            front_layer (list): front layer gates list
            extended_set: set of expansion gates
            unavailable_2qubits (set): set of unavailable two-qubits
        Returns:
            best_swap (tuple): the best swap based on different heuristics
        """
        swap_scores = dict.fromkeys(swap_candidates, -100000)
        if self.heuristic == 'fidelity':
            for swap_qubits in swap_scores:
                physical_swap_qubits = (current_layout.v2p[swap_qubits[0]], current_layout.v2p[swap_qubits[1]])
                # Determine whether the added swap gate is valid, because swap(1,2)swap(1,2)=identity
                if tuple(sorted(physical_swap_qubits)) not in unavailable_2qubits:
                    swap_cost = self._swap_score(physical_swap_qubits)
                    score_h = self._score_heuristic(
                        self.heuristic, front_layer, extended_set, current_layout, swap_qubits
                    )
                    score = swap_cost + score_h
                    swap_scores[swap_qubits] = score

            best_score = max(swap_scores.values())
            best_swaps = [k for k, v in swap_scores.items() if v == best_score]
            if len(best_swaps) == 1:
                best_swap = best_swaps[0]
            else:
                best_swap = choice(best_swaps)  # randomly select one

        elif self.heuristic == 'distance':
            for swap_qubits in swap_scores:
                score = self._score_heuristic(
                    self.heuristic, front_layer, extended_set, current_layout, swap_qubits
                )
                swap_scores[swap_qubits] = score

            best_score = min(swap_scores.values())
            best_swaps = [k for k, v in swap_scores.items() if v == best_score]
            if len(best_swaps) == 1:
                best_swap = best_swaps[0]
            else:
                best_swap = choice(best_swaps)  # randomly select one

        elif self.heuristic == 'mixture':
            for swap_qubits in swap_scores:
                score = self._score_heuristic(
                    'distance', front_layer, extended_set, current_layout, swap_qubits
                )
                swap_scores[swap_qubits] = score

            best_score = min(swap_scores.values())
            best_swaps = [k for k, v in swap_scores.items() if v == best_score]
            best_swap = best_swaps[0]

            if len(best_swaps) > 1:
                max_score = 0
                for swap_qubits in best_swaps:
                    physical_swap_qubits = (current_layout.v2p[swap_qubits[0]], current_layout.v2p[swap_qubits[1]])
                    if tuple(sorted(physical_swap_qubits)) not in unavailable_2qubits:
                        swap_cost = self._swap_score(physical_swap_qubits)
                        score_h = self._score_heuristic(
                            'fidelity', front_layer, extended_set, current_layout, swap_qubits
                        )
                        score = swap_cost + score_h
                        if max_score > score:
                            max_score = score
                            best_swap = swap_qubits

        else:
            raise NameError("No heuristic %s, currently only 'distance' and 'fidelity' are supported." % self.heuristic)
        return best_swap

    def _compute_distance_cost(self, layer, layout):
        """Calculate the distance sum of the gates list (layer) of the given layout.

        Args:
            layer: Gates list of a certain layer of the circuit.
            layout (Layout): Mapping of virtual qubits to physical qubits.

        Returns:
            cost (int): distance sum of the gates list in the given layer.
        """
        cost = 0
        for node in layer:
            cost += self.distance_matrix[layout.v2p[node.pos[0]], layout.v2p[node.pos[1]]]
        return cost

    def _compute_fidelity_cost(self, layer, layout):
        """Calculate the fidelity sum of the gates list (layer) of the given layout.

        Args:
            layer (list): Gates list of a certain layer of the circuit.
            layout (Layout): Mapping of virtual qubits to physical qubits.

        Returns:
            fidelity_cost (int): path_fidelity sum of the gates list in the given layer.
                                Since there may be exchanges of target qubit and control qubit that make
                                the fidelity of the two-qubit gate different, the average is taken here.
        """
        fidelity_cost = 0
        for node in layer:
            p1, p2 = layout.v2p[node.pos[0]], layout.v2p[node.pos[1]]
            fidelity_cost += 0.5 * (self.path_fidelity[(p1, p2)] + self.path_fidelity[(p2, p1)])
        return fidelity_cost

    def _compute_mixture_cost(self, layer, layout):
        """
        TODO: Deprecated
        """
        mixture_cost = 0
        sum_fidelity = 0
        for node in layer:
            s1, s2 = layout.v2p[node.pos[0]], layout.v2p[node.pos[1]]
            mean_fidelity_s12 = 0.5 * (np.exp(self.path_fidelity[(s1, s2)]) + np.exp(self.path_fidelity[(s2, s1)]))
            mixture_cost += self.distance_matrix[s1, s2] * mean_fidelity_s12
            sum_fidelity += mean_fidelity_s12
        mixture_cost = mixture_cost / sum_fidelity
        # mixture_cost = mixture_cost
        return mixture_cost

    def _score_heuristic(self, heuristic, front_layer, extended_set, current_layout, swap_qubits=None):
        """Get the heuristic score for the current layout.

        Args:
            heuristic (str): "distance" or "fidelity"
            front_layer (list): front layer gates list
            extended_set: set of expansion gates
            current_layout (Layout): current layout
            swap_qubits (tuple): swap qubits (i, j)

        Returns:
            the heuristic score
        """
        # trial_layout = copy.deepcopy(current_layout)
        trial_layout = Layout()
        trial_layout.set_layout(current_layout)
        trial_layout.swap(swap_qubits[0], swap_qubits[1])
        if heuristic == "distance":
            front_cost = self._compute_distance_cost(front_layer, trial_layout) / len(front_layer)
            extended_cost = 0
            if extended_set:
                extended_cost = self._compute_distance_cost(extended_set, trial_layout) / len(extended_set)
            total_cost = front_cost + self.extended_set_weight * extended_cost
            return total_cost * max(self.qubits_decay[swap_qubits[0]], self.qubits_decay[swap_qubits[1]])

        elif heuristic == "fidelity":
            # noise_front_cost = self._compute_fidelity_cost(front_layer, layout) / len(front_layer)
            noise_front_cost = self._compute_fidelity_cost(front_layer, trial_layout)
            noise_extended_cost = 0
            if extended_set:
                # noise_extended_cost = self._compute_fidelity_cost(extended_set, layout) / len(extended_set)
                noise_extended_cost = self._compute_fidelity_cost(extended_set, trial_layout)
            noise_total_cost = noise_front_cost + self.extended_set_weight * noise_extended_cost
            return 0.5 * (self.qubits_decay[swap_qubits[0]] + self.qubits_decay[swap_qubits[1]]) * noise_total_cost

        # elif heuristic == "mixture":
        #     mixture_front_cost = self._compute_mixture_cost(front_layer, trial_layout)
        #     mixture_extended_cost = 0
        #     if extended_set:
        #         mixture_extended_cost = self._compute_mixture_cost(extended_set, trial_layout)
        #     mixture_total_cost = mixture_front_cost + self.extended_set_weight * mixture_extended_cost
        #     return 0.5 * (self.qubits_decay[swap_qubits[0]] + self.qubits_decay[swap_qubits[1]]) * mixture_total_cost

        else:
            raise NameError("Heuristic %s not recognized." % heuristic)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qsteed.passes.mapping.layout.sabre_layout import SabreLayout
from qsteed.passes.unroll.unroll_to_2qubit import UnrollTo2Qubit
from qsteed.passes.unroll.unroll_to_basis import UnrollToBasis
from qsteed.passes.optimization.optimization_combine import GateCombineOptimization
from qsteed.passes.optimization.one_qubit_optimization import OneQubitGateOptimization
from .passflow import PassFlow


def level_0_passflow(basis_gates):
    level_0 = [UnrollTo2Qubit(),
               UnrollToBasis(basis_gates=basis_gates)
               ]
    return PassFlow(level_0)


def level_1_passflow(basis_gates):
    level_1 = [UnrollTo2Qubit(),
               SabreLayout(heuristic='distance', max_iterations=3),
               UnrollToBasis(basis_gates=basis_gates),
               GateCombineOptimization(),
               OneQubitGateOptimization(),
               ]
    return PassFlow(level_1)


def level_2_passflow(basis_gates):
    level_2 = [UnrollTo2Qubit(),
               SabreLayout(heuristic='fidelity', max_iterations=3),
               UnrollToBasis(basis_gates=basis_gates),
               GateCombineOptimization(),
               OneQubitGateOptimization(),
               ]
    return PassFlow(level_2)


def level_3_passflow(basis_gates):
    level_3 = [UnrollTo2Qubit(),
               SabreLayout(heuristic='mixture', max_iterations=3),
               UnrollToBasis(basis_gates=basis_gates),
               GateCombineOptimization(),
               OneQubitGateOptimization(),
               ]
    return PassFlow(level_3)


class PresetPassflow(PassFlow):
    """A qsteed passflow describes a quantum circuit compilation process."""

    def __init__(self, basis_gates, optimization_level=0):
        super().__init__()
        self.basis_gates = basis_gates
        self.optimization_level = optimization_level

    def get_passflow(self):
        if self.optimization_level == 0:
            return level_0_passflow(self.basis_gates)
        elif self.optimization_level == 1:
            return level_1_passflow(self.basis_gates)
        elif self.optimization_level == 2:
            return level_2_passflow(self.basis_gates)
        elif self.optimization_level == 3:
            return level_3_passflow(self.basis_gates)
        else:
            raise ValueError("Error: The value of optimization_level is between [0,3].")



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Here are some examples of functions commonly used in tests"""

from qsteed.backends.backend import Backend
from qsteed.passes.mapping.layout.sabre_layout import SabreLayout
from qsteed.passes.model import Model
from qsteed.passes.optimization.one_qubit_optimization import OneQubitGateOptimization
from qsteed.passes.optimization.optimization_combine import GateCombineOptimization
from qsteed.passes.unroll.unroll_to_2qubit import UnrollTo2Qubit
from qsteed.passes.unroll.unroll_to_basis import UnrollToBasis
from qsteed.passflow.passflow import PassFlow
from qsteed.utils.random_circuit import RandomCircuit

# Test unified parameter settings
coupling_list = [(2, 3, 0.982), (3, 2, 0.982), (3, 4, 0.973), (4, 3, 0.973),
                 (0, 1, 0.98), (1, 0, 0.98), (1, 2, 0.97), (2, 1, 0.97)]  # qubit numbering starts from 0
qubits = 5  # The number of qubits in the random circuit should match the number of qubits in the chip's coupling_list.
# Specify the basic gate set that the transpiled circuit will ultimately use.
basis_gates = ['cx', 'rx', 'ry', 'rz', 'id', 'h']
# A set of quantum gates used to generate random circuit. Set to None to use all supported quantum gates.
gates_list = ['ccx', 'rxx', 'rz', 'ry', 'h', 'x', 'y', 'z']


def get_random_circuit(gates_number: int = 20, depth: int = None):
    """Only one parameter, either the gates_number or the depth, needs to be specified,
        with preference given to the gates_number if both are provided."""
    rqc = RandomCircuit(num_qubit=qubits, gates_number=gates_number, depth=depth,
                        gates_list=gates_list, measure=True)
    qc = rqc.random_circuit()
    return qc


def get_backend_properties():
    backend_properties = {
        'name': 'ExampleBackend',
        'backend_type': 'superconducting',
        'qubits_num': qubits,
        'coupling_list': coupling_list,
        'basis_gates': basis_gates,
    }
    return backend_properties


def get_initial_model():
    backend_properties = get_backend_properties()
    backend_instance = Backend(**backend_properties)
    initial_model = Model(backend=backend_instance)
    return initial_model


def get_passflow():
    backend_properties = get_backend_properties()
    passes = [
        UnrollTo2Qubit(),
        SabreLayout(heuristic='fidelity', max_iterations=3),  # heuristic='distance' or 'fidelity', 'mixture'
        UnrollToBasis(basis_gates=backend_properties['basis_gates']),
        GateCombineOptimization(),
        OneQubitGateOptimization(),
    ]
    passflow = PassFlow(passes=passes)
    return passflow



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import math
import re
from queue import PriorityQueue

import networkx as nx


class BuildLibrary:
    """Based on the hardware coupling structure and the fidelity information of two-qubit gates,
    construct an optimal qubit coupling substructure library."""

    def __init__(self, backend: str = 'ScQ-P10', fidelity_threshold: float = 0.9):
        """Initialize BuildLibrary.
        Args:
            backend: quafu backends ('ScQ-P10', 'ScQ-P18', 'ScQ-P136').
            fidelity_threshold: Two-qubit gate fidelity.
            backend_info (dict): Get backend information from :py:func:`quafu.Task.get_backend_info()`
        """
        self.backend = backend
        self.fidelity_threshold = fidelity_threshold

    def get_structure(self, json_topo_struct):
        """
        Args:
            json_topo_struct(dict): backend_info['full_info']["topological_structure"]
        Returns:
            int_to_qubit(dict): {0: Q01, 1: Q02, ...}
            qubit_to_int(dict): {Q01: 0, Q02: 0, ...}
            directed_weighted_edges(list):[[qubit1,qubit2,fidelity],...]
            connected_substructure_list(list): [networkx.Graph,...]
        """
        # Please note that 'int_to_qubit' needs to be consistent with the mapping of the backend,
        # and you can check 'backend_info['mapping']'.

        qubits_list = []
        for gate in json_topo_struct.keys():
            qubit = gate.split('_')
            qubits_list.append(qubit[0])
            qubits_list.append(qubit[1])
        qubits_list = list(set(qubits_list))
        qubits_list = sorted(qubits_list, key=lambda x: int(re.findall(r"\d+", x)[0]))
        int_to_qubit = {k: v for k, v in enumerate(qubits_list)}
        qubit_to_int = {v: k for k, v in enumerate(qubits_list)}

        directed_weighted_edges = []
        for gate, name_fidelity in json_topo_struct.items():
            gate_qubit = gate.split('_')
            qubit1 = qubit_to_int[gate_qubit[0]]
            qubit2 = qubit_to_int[gate_qubit[1]]
            gate_name = list(name_fidelity.keys())[0]
            fidelity = name_fidelity[gate_name]['fidelity']
            directed_weighted_edges.append([qubit1, qubit2, fidelity])

        connected_substructure_list = self.connected_substructure(directed_weighted_edges)
        return int_to_qubit, qubit_to_int, directed_weighted_edges, connected_substructure_list

    def connected_substructure(self, directed_weighted_edges: list = None):
        d = {}
        for item in directed_weighted_edges:
            key = tuple(sorted(item[:2]))
            if key in d:
                if item[2] > d[key][2]:
                    d[key] = item
            else:
                d[key] = item
        weighted_edges = list(d.values())
        G = nx.Graph()
        G.add_weighted_edges_from(weighted_edges)
        G0 = copy.deepcopy(G)
        connected_substructure_list = []
        while len(G.nodes()) > 0:
            connected_nodes = max(nx.connected_components(G), key=len)
            connected_subgraph = G0.subgraph(connected_nodes)
            connected_substructure_list.append(connected_subgraph)
            G.remove_nodes_from(connected_nodes)
        return connected_substructure_list

    def substructure(self, directed_weighted_edges, connected_substructure_list, qubits_need):
        """
        Args:
            directed_weighted_edges(list): [[qubit1,qubit2,fidelity],...], get from get_structure()
            connected_substructure_list(list): [networkx.Graph,...], get from get_structure()
            qubits_need: number of qubits required
        Returns:
            all_substructure(list): All coupling substructures with 'qubits_need' qubits
        """
        if directed_weighted_edges[0][2] > 1:
            directed_weighted_edges = [[item[0], item[1], item[2] / 100] for item in directed_weighted_edges]
        fidelity_threshold_fixed = self.fidelity_threshold
        all_substructure = []
        # substructure_nodes = []
        while len(all_substructure) == 0:
            for cg in connected_substructure_list:
                if len(cg.nodes()) >= qubits_need:
                    substructure_nodes = []  #
                    fidelity_threshold = fidelity_threshold_fixed  #
                    sorted_edges = sorted(cg.edges(data=True), key=lambda x: x[2]['weight'], reverse=True)
                    for elem in sorted_edges:
                        if elem[2]['weight'] > fidelity_threshold:
                            neighbors = PriorityQueue()
                            neighbors.put((-1, elem[0]))
                            ret_nodes = []
                            log_weight_product = 0
                            for node in cg.nodes():
                                cg.nodes[node]['visited'] = False
                            while neighbors.not_empty:
                                temp = neighbors.get()
                                node = temp[1]
                                if cg.nodes[node]['visited']:
                                    continue
                                weight = -temp[0]
                                if weight <= 0:
                                    weight = 1e-10
                                log_weight_product += math.log(weight)
                                cg.nodes[node]['visited'] = True
                                ret_nodes.append(node)
                                if len(ret_nodes) == qubits_need:
                                    break
                                for neighbor in cg[node]:
                                    if not cg.nodes[neighbor]['visited']:
                                        weight = cg[node][neighbor]['weight']
                                        neighbors.put((-weight, neighbor))
                            out = []
                            for edge in directed_weighted_edges:
                                if edge[0] in ret_nodes and edge[1] in ret_nodes:
                                    out.append(edge)
                            if sorted(ret_nodes) not in substructure_nodes and all(
                                    qubit[2] > fidelity_threshold for qubit in out):
                                substructure_nodes.append(sorted(ret_nodes))
                                # all_substructure.append([log_weight_product, out])
                                ave_weight = 0  # Dealing with loop structures
                                for e in out:
                                    if e[2] <= 0:  # If fidelity is 0, set it to 1e-10
                                        e[2] = 1e-10
                                    ave_weight += math.log(e[2])
                                ave_weight = ave_weight / len(out)
                                all_substructure.append([ave_weight, out])
                            if not substructure_nodes:
                                fidelity_threshold = fidelity_threshold - 0.01
            # If the current threshold does not find a substructure, reduce the threshold value.
            if len(all_substructure) == 0:
                fidelity_threshold_fixed = fidelity_threshold_fixed - 0.01
        all_substructure = sorted(all_substructure, key=lambda x: x[0], reverse=True)
        return all_substructure

    def build_substructure_library(self, json_topo_struct):
        # substructure_dict = defaultdict(list)
        substructure_dict = {}
        int_to_qubit, qubit_to_int, directed_weighted_edges, connected_substructure_list = self.get_structure(
            json_topo_struct)
        substructure_dict[1] = [[[q, q, 0.99]] for q in int_to_qubit.keys()]
        for qubits in range(2, len(connected_substructure_list[0].nodes()) + 1):
            sub_graph = self.substructure(directed_weighted_edges, connected_substructure_list, qubits)
            qlist = []
            for j in range(len(sub_graph)):
                sublist = sub_graph[j][1]
                sublist.sort(key=lambda x: x[0])
                qlist.append(sublist)
            substructure_dict[qubits] = qlist

        sorted_weighted_edges = sorted(directed_weighted_edges, key=lambda x: x[2], reverse=True)
        save_substructure = {'structure': sorted_weighted_edges, 'substructure_dict': substructure_dict,
                             'int_to_qubit': int_to_qubit, 'qubit_to_int': qubit_to_int}
        return save_substructure



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import numpy as np
from quafu.elements import Instruction
from quafu.elements.element_gates import ControlledU, CXGate, RZGate, RYGate
from scipy.linalg import sqrtm

from qsteed.passes.basepass import UnrollPass
from qsteed.passes.decomposition.ZYZ_decompose import zyz_decomposition


class ControlledUToCNOT(UnrollPass):
    '''The ControllUToCNOT
    q[0]  --*--      q[0]  ---------------------------------+----*-----+----*----+-----*----+----*--------
            |                                               |    |     |    |    |     |    |    |
    q[1]  --*--  ≡   q[1]  -------------+----*----+----*----*----|-----|----|----*-----|----|----|--------
            |                           |    |    |    |         |     |    |          |    |    |
    q[2]  --*--      q[2]  -------*-----*----|----*----|---------|-----*----|----------|----*----|--------
            |                     |          |         |         |          |          |         |  
    q[3]  --X--      q[3]  -------V----------V---------V---------V----------V----------V---------V------- 


    q[0]  --*--      q[0]   --------------*--------------------   
            |    ≡                        |
    q[1]  --V--      q[1]   --RZ(gamma)---+------RY(beta)------


    q[0]  --*--     q[0]    -------*---------*----*---------
            |                      |         |    |
    q[1]  --*-- ≡   q[1]    --*----+----*----+----|---------
            |                 |         |         |
    q[2]  --V--     q[2]    --V---------V*--------V---------
        '''

    def __init__(self) -> None:
        super().__init__()
        self.original = ControlledU.name.lower()
        self.basis = [CXGate.name.lower(), RZGate.name.lower(), RYGate.name.lower()]

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, ControlledU):
            control_bits = op.ctrls
            target_bit = op.targs[0]
            U = op.targ_gate
            if len(control_bits) == 1:
                gamma, beta, alpha, global_phase = zyz_decomposition(U)
                rule.append(RZGate(target_bit, gamma))
                rule.append(CXGate(control_bits[0], target_bit))
                rule.append(RYGate(target_bit, beta))
            elif len(control_bits) == 2:
                V = self.decompose_U(U)
                gamma1, beta1, alpha1, global_phase1 = zyz_decomposition(V)
                gamma2, beta2, alpha2, global_phase2 = zyz_decomposition(np.conj(V).T)
                rule.append(RZGate(target_bit, gamma1))
                rule.append(CXGate(control_bits[1], target_bit))
                rule.append(RYGate(target_bit, beta1))
                rule.append(CXGate(control_bits[0], control_bits[1]))
                rule.append(RZGate(target_bit, gamma2))
                rule.append(CXGate(control_bits[1], target_bit))
                rule.append(RYGate(target_bit, beta2))
                rule.append(CXGate(control_bits[0], control_bits[1]))
                rule.append(RZGate(target_bit, gamma1))
                rule.append(CXGate(control_bits[0], target_bit))
                rule.append(RYGate(target_bit, beta1))
            else:
                graycode = self.build_gray_code(len(control_bits))
                last_code = graycode[0, :]
                decompose_times = len(control_bits - 1)
                V = U
                for _ in decompose_times:
                    V = self.decompose_U(U)
                gamma1, beta1, alpha1, global_phase1 = zyz_decomposition(V)

                gamma2, beta2, alpha2, global_phase2 = zyz_decomposition(np.conj(V).T)
                for i in range(1, len(graycode)):
                    code = graycode[i, :]
                    ones_bits = np.where(code == 1)[0]
                    set_idx = ones_bits[0]
                    diff_idx = np.where(code != last_code)[0][0]
                    if diff_idx != set_idx:
                        rule.append(CXGate(control_bits[diff_idx], control_bits[set_idx]))
                    elif len(ones_bits) >= 2:
                        next_idx = ones_bits[1]
                        rule.append(CXGate(control_bits[next_idx], control_bits[set_idx]))
                    if np.sum(code) % 2 == 0:
                        rule.append(RZGate(target_bit, gamma2))
                        rule.append(CXGate(control_bits[set_idx], target_bit))
                        rule.append(RYGate(target_bit, beta2))
                    else:
                        rule.append(RZGate(target_bit, gamma1))
                        rule.append(CXGate(control_bits[set_idx], target_bit))
                        rule.append(RYGate(target_bit, beta1))
                    last_code = code

        else:
            rule.append(op)
        self.rule = rule
        return rule

    # build 2^n gray code based on the control bits
    def build_gray_code(self, n):
        code = np.full((2 ** n, n), 0)
        block = np.repeat([0, 1], 2 ** (n - 1))
        code[:, 0] = block
        for i in range(1, n):
            block = np.repeat([0, 1, 1, 0], 2 ** (n - 1 - i))
            print(block)
            code[:, i] = np.tile(block, 2 ** (i - 1))
        return code

    # decompose U to V, where U = V^2
    def decompose_U(self, U):
        V = sqrtm(U)
        return V



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from qsteed.dag.circuit_dag_convert import circuit_to_dag, dag_to_circuit, draw_dag, nodelist_to_dag, \
    gate_to_node
from tests.shared_utils import get_random_circuit


class TestDAGConversion:

    def test_circuit_to_dag_and_back(self):
        qc = get_random_circuit(gates_number=20)
        assert qc is not None
        qc.draw_circuit()

        dag = circuit_to_dag(qc)
        # draw_dag(dag)  # You can uncomment the lines locally to display the DAG diagram.
        # show_dag(dag)

        re_qc = dag_to_circuit(dag, qubits=qc.num)
        assert re_qc is not None
        re_qc.draw_circuit()
        assert len(qc.gates) == len(re_qc.gates), "Number of gates should be the same"

    def test_nodelist_to_dag(self):
        qc = get_random_circuit(gates_number=20)
        assert qc is not None
        qc.draw_circuit()

        nodes_list = [gate_to_node(gate, specific_label=i) for i, gate in enumerate(qc.gates)]
        dag = nodelist_to_dag(nodes_list)
        # draw_dag(dag)  # You can uncomment the lines locally to display the DAG diagram.
        # show_dag(dag)

        re_qc = dag_to_circuit(dag, qubits=qc.num)
        assert re_qc is not None
        assert len(qc.gates) == len(re_qc.gates), "Number of gates should be the same"



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import pi
from typing import List

import numpy as np
from quafu.elements import Instruction
from quafu.elements.element_gates import MCYGate, CXGate, ToffoliGate, CPGate, SdgGate, SGate, HGate, CYGate

from qsteed.passes.basepass import UnrollPass


class MCYToCNOT(UnrollPass):
    """The MCYToCNOT pass.
    convert MCY gate to {CNOT, H, S,Sdg,CP}.

    MCY gate decomposition rule(control bit 0:2, target bit: 3, ancilla bit: None):
    q[0]  --*--      q[0]  ---------------------------------------+----*----+----*----+----*----+----*--------
            |                                                     |    |    |    |    |    |    |    |
    q[1]  --*--  ≡   q[1]  -------------------+----*----+----*----*----|----|----|----*----|----|----|--------
            |                                 |    |    |    |         |    |    |         |    |    |
    q[2]  --*--      q[2]  --------------*----*----|----*----|---------|----*----|---------|----*----|--------
            |                            |         |         |         |         |         |         |  
    q[3]  --X--      q[3]  --Sdg----H----R1--------R1--------R1--------R1--------R1--------R1--------R1----H----S-- 
    
    """

    def __init__(self) -> None:
        super().__init__()
        self.original = MCYGate([0, 1], 2).name.lower()
        self.basis = [CXGate.name.lower(), SdgGate.name.lower(), SGate.name.lower(), CPGate(0, 1, 0).name.lower(),
                      CYGate.name.lower(), ToffoliGate(0, 1, 2).name.lower()]

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, MCYGate):
            control_bits = op.ctrls
            if isinstance(op.targs, list):
                target_bit = op.targs[0]
            else:
                target_bit = op.targs

            if len(control_bits) == 1:
                rule.append(CYGate(control_bits[0], target_bit))
            elif len(control_bits) == 2:
                rule.append(SdgGate(target_bit))
                rule.append(ToffoliGate(control_bits[0], control_bits[1], target_bit))
                rule.append(SGate(target_bit))
            else:
                graycode = self.build_gray_code(len(control_bits))
                theta = pi / 2 ** (len(control_bits) - 1)
                last_code = graycode[0, :]
                rule.append(SdgGate(target_bit))
                rule.append(HGate(target_bit))

                for i in range(1, len(graycode)):
                    code = graycode[i, :]
                    ones_bits = np.where(code == 1)[0]
                    set_idx = ones_bits[0]
                    diff_idx = np.where(code != last_code)[0][0]
                    if diff_idx != set_idx:
                        rule.append(CXGate(control_bits[diff_idx], control_bits[set_idx]))
                    elif len(ones_bits) >= 2:
                        next_idx = ones_bits[1]
                        rule.append(CXGate(control_bits[next_idx], control_bits[set_idx]))
                    if np.sum(code) % 2 == 0:
                        rule.append(CPGate(control_bits[set_idx], target_bit, -theta))
                    else:
                        rule.append(CPGate(control_bits[set_idx], target_bit, theta))
                    last_code = code

                rule.append(HGate(target_bit))
                rule.append(SGate(target_bit))
        else:
            rule.append(op)
        self.rule = rule
        return rule
        # build graycode based on the control bits

    def build_gray_code(self, n):
        code = np.full((2 ** n, n), 0)
        block = np.repeat([0, 1], 2 ** (n - 1))
        code[:, 0] = block
        for i in range(1, n):
            block = np.repeat([0, 1, 1, 0], 2 ** (n - 1 - i))
            print(block)
            code[:, i] = np.tile(block, 2 ** (i - 1))
        return code



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from quafu.elements import Instruction
from quafu.elements.element_gates import U3Gate
from quafu.elements.element_gates.rotation import RXGate, RYGate, RZGate

from qsteed.passes.basepass import UnrollPass
from qsteed.passes.decomposition.unitary_decompose import UnitaryDecompose


class U3Decompose(UnrollPass):
    """The U3Decompose pass.
    convert U3 gate to {Rx, RY, RZ}.

    U3 gate decomposition rule:
    U3 decomposition method: ZYZ or ZXZ or XYX or XZX
    """

    def __init__(self, one_qubit_decompose: str = 'ZXZ') -> None:
        super().__init__()
        self.original = U3Gate(0, 0, 0, 0).name.lower()
        self.basis = [RXGate(0, 0).name.lower(), RYGate(0, 0).name.lower(), RZGate(0, 0).name.lower()]
        self.one_qubit_decompose = one_qubit_decompose

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, U3Gate):
            if isinstance(op.pos, list):
                pos = op.pos[0]
            else:
                pos = op.pos

            uc = UnitaryDecompose(op.matrix, [0], one_qubit_decompose=self.one_qubit_decompose)
            uc.decompose()
            gates = uc.quafuQC.gates

            for g in gates:
                g.pos = [pos]
                rule.append(g)
        else:
            rule.append(op)
        self.rule = rule
        return rule



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict

from qsteed.backends.backend import Backend
from qsteed.passes.datadict import DataDict


class Model:
    """
    Represents a model of a quantum system.

    Attributes:
        _backend (Backend): The backend of the quantum system.
        _layout (List[int]): Layout of the quantum system.
    """

    def __init__(self, backend: Backend, layout: Dict = None, datadict: DataDict = None):
        self._backend = backend
        self._layout = {'initial_layout': None, 'final_layout': None}
        self.set_layout(layout)
        self.datadict = datadict

    def get_backend(self) -> Backend:
        """Return the backend of the model."""
        return self._backend

    # def set_layout(self, layout: List[int]):
    #     """Set a new layout for the quantum system."""
    #     self._layout = layout

    def set_layout(self, new_layout: Dict):
        """Set a new layout for the quantum system."""
        if isinstance(new_layout, Dict):
            if 'initial_layout' in new_layout.keys():
                self._layout['initial_layout'] = new_layout['initial_layout']
            if 'final_layout' in new_layout.keys():
                self._layout['final_layout'] = new_layout['final_layout']

    def get_layout(self) -> Dict:
        """Return the layout of the quantum system."""
        return self._layout

    def get_datadict(self) -> Dict:
        """Return the datadict of the quantum system."""
        return self.datadict

    def __repr__(self):
        return f"<Model(backend={self.get_backend()}, layout={self.get_layout()})>"

    def backend_info(self) -> str:
        """Return a string representation of the backend's information."""
        return repr(self._backend)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing
import os

import dill

from qsteed.passes.model import Model
from qsteed.passflow.passflow import PassFlow
from qsteed.transpiler.transpiler import Transpiler


def process_circuit(serialized_circuit, passflow, model):
    circuit = dill.loads(serialized_circuit)
    transpiler = Transpiler(passflow, model)
    compiled_circuit = transpiler.transpile(circuit)
    serialized_compiled_circuit = dill.dumps(compiled_circuit)
    return serialized_compiled_circuit


def parallel_process_circuits(circuits, passflows, models, num_processes=None):
    if isinstance(passflows, list):
        if len(passflows) != len(circuits):
            raise ValueError("The length of circuits and passflows are not equal.")
    elif isinstance(passflows, PassFlow):
        passflows = [passflows for _ in range(len(circuits))]
    else:
        raise ValueError(
            "Please enter the correct passflows data type, a single passflow or a list of multiple passflows.")

    if isinstance(models, list):
        if len(models) != len(circuits):
            raise ValueError("The length of circuits and models are not equal.")
    elif isinstance(models, Model):
        models = [models for _ in range(len(circuits))]
    else:
        raise ValueError(
            "Please enter the correct models data type, a single model or a list of multiple models.")

    serialized_circuits = [dill.dumps(circuit) for circuit in circuits]

    picked = zip(serialized_circuits, passflows, models)

    if num_processes is None:
        cpu_count = os.cpu_count()
        half_cpu_count = int(cpu_count / 2)  # Half full load
        if len(circuits) > half_cpu_count:
            num_processes = half_cpu_count
        else:
            num_processes = len(circuits)

    with multiprocessing.Pool(num_processes) as pool:
        results = pool.starmap(process_circuit, [(circuit, passflow, model) for circuit, passflow, model in picked])

    compiled_circuits = []
    for qc in results:
        compiled_circuit = dill.loads(qc)
        compiled_circuits.append(compiled_circuit)
    return compiled_circuits



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random


class Layout:
    """Layout between virtual qubits and physical qubits."""

    def __init__(self, mapping: dict = None):
        """mapping virtual qubits to physical qubits

        Args:
            mapping (dict): {virtual qubit: physical qubit, v0: p0, v1: p1, ...}
        """
        if mapping is not None:
            self.v2p = mapping
            self.p2v = {p: v for v, p in mapping.items()}
        else:
            self.p2v = {}
            self.v2p = {}

    def set_layout(self, layout):
        self.v2p = layout.v2p.copy()
        self.p2v = layout.p2v.copy()

    def generate_trivial_layout(self, virtual_qubits: int = None):
        """
        Args:
            virtual_qubits (int): the number of virtual qubits in a circuit
        """
        self.v2p = {k: k for k in range(virtual_qubits)}
        self.p2v = {k: k for k in range(virtual_qubits)}

    def generate_random_layout(self, virtual_qubits: int = None, physical_qubits: int = None):
        """
        Args:
            virtual_qubits (int): the number of virtual qubits in a circuit
            physical_qubits (int): the number of physical qubits in a chip

        raise: The number of virtual qubits in the circuit cannot be greater than
                the number of physical qubits in the chip.

        """
        if virtual_qubits <= physical_qubits:
            virtual_qubit = random.sample(range(virtual_qubits), virtual_qubits)
            physical_qubit = random.sample(range(physical_qubits), virtual_qubits)
            for v, q in zip(virtual_qubit, physical_qubit):
                self.v2p[v] = q
            self.p2v = {p: v for v, p in self.v2p.items()}
        else:
            raise ValueError('Error: The number of virtual qubits in the circuit cannot be greater than '
                             'the number of physical qubits in the chip.')

    def from_v2p_dict(self, input_dict: dict = None):
        """
        Args:
            input_dict (dict): {virtual qubit: physical qubit, v0: p0, v1: p1, ...}
        """
        if input_dict is not None:
            self.v2p = input_dict
            self.p2v = {p: v for v, p in input_dict.items()}

    def from_p2v_dict(self, input_dict: dict = None):
        """
        Args:
            input_dict (dict): {physical qubit: virtual qubit, p0: v0, p1: v1, ...}
        """
        if input_dict is not None:
            self.p2v = input_dict
            self.v2p = {v: p for p, v in input_dict.items()}

    def swap(self, a, b):
        self.v2p[a], self.v2p[b] = self.v2p[b], self.v2p[a]
        self.p2v = {p: v for v, p in self.v2p.items()}



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from quafu import QuantumCircuit


def reverse_circuit(circuit: QuantumCircuit):
    """
    This function aims to provide a reversed circuit for a given circuit.

    Args:
        circuit: input circuit that need to be reversed
    Return:
        the reversed circuit aimed to do bidirectional routing
    """
    rev_circuit = QuantumCircuit(circuit.num)

    if circuit.measures:
        for qubit, cbit in circuit.measures.items():
            rev_circuit.measure([qubit], [cbit])

    for gate in circuit.gates[::-1]:
        rev_circuit.add_gate(gate)
    return rev_circuit



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from scipy.stats import unitary_group

from qsteed.passes.decomposition.unitary_decompose import UnitaryDecompose
from qsteed.passes.decomposition.utils import convert_utils as cu
from qsteed.passes.decomposition.utils import matrix_utils as mu


class TestUnitaryDecompose:
    def test_unitary_decompose(self):
        # Create random unitary
        qubits = 4
        U0 = unitary_group.rvs(2 ** qubits)

        # Decompose any unitary matrix to obtain quantum circuit
        qubit_list = list(range(qubits))
        ud = UnitaryDecompose(U0, qubit_list, one_qubit_decompose='XZX')
        ud.decompose()
        gates_list = ud.gates_list
        quafuQC = ud.quafuQC
        quafuQC.draw_circuit()

        # Quantum circuit conversion to unitary matrix
        quafuU = cu.circuit_to_unitary(quafuQC)
        assert mu.is_approx(quafuU * np.exp(1j * ud.global_phase),
                            U0), "The decomposed circuit is not consistent with the original unitary"
        print("Is the decomposed circuit consistent with the original unitary?",
              mu.is_approx(quafuU * np.exp(1j * ud.global_phase), U0))

        # Quantum gates list conversion to unitary matrix
        du = cu.gates_list_to_unitary(qubits, gates_list)
        assert mu.is_approx(du * np.exp(1j * ud.global_phase),
                            U0), "The decomposed circuit is not consistent with the original unitary"
        print("Is the decomposed circuit consistent with the original unitary?",
              mu.is_approx(du * np.exp(1j * ud.global_phase), U0))

        # Calculate the distance between the original and decomposed unitary
        distance = mu.matrix_distance_squared(quafuU, U0)
        assert distance < 1e-9, f"The distance between the original and decomposed unitary is too large: {distance}"
        print('The distance between the original and decomposed unitary:', distance)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Any

import networkx as nx
import numpy as np
from quafu import QuantumCircuit
from quafu.elements import Barrier, Delay, XYResonance
from quafu.elements.element_gates import *
from quafu.elements.element_gates.clifford import *
from quafu.elements.element_gates.pauli import *

from .dagcircuit import DAGCircuit
from .instruction_node import InstructionNode

GATE_CLASSES = {
    "x": XGate,
    "y": YGate,
    "z": ZGate,
    "h": HGate,
    "s": SGate,
    "sdg": SdgGate,
    "t": TGate,
    "tdg": TdgGate,
    "rx": RXGate,
    "ry": RYGate,
    "rz": RZGate,
    "id": IdGate,
    "sx": SXGate,
    "sxdg": SXdgGate,
    "sy": SYGate,
    "sydg": SYdgGate,
    "w": WGate,
    "sw": SWGate,
    "swdg": SWdgGate,
    "p": PhaseGate,
    "delay": Delay,
    "barrier": Barrier,
    "cx": CXGate,
    "cp": CPGate,
    "swap": SwapGate,
    "iswap": ISwapGate,
    "rxx": RXXGate,
    "ryy": RYYGate,
    "rzz": RZZGate,
    "cy": CYGate,
    "cz": CZGate,
    "cs": CSGate,
    "ct": CTGate,
    "cry": CRYGate,
    "crx": CRXGate,
    "crz": CRZGate,
    "xy": XYResonance,
    "ccx": ToffoliGate,
    "cswap": FredkinGate,
    "mcx": MCXGate,
    "mcy": MCYGate,
    "mcz": MCZGate,
    "u3": U3Gate,
}


def gate_to_node(input_gate, specific_label):
    """
    Transform a gate in a quantum circuit of quafu (excluding measure gates)
    into a node in the graph with a specific label.

    Args:
        input_gate: a gate in quantumcircuit of quafu(not include measure_gate)
        specific_label: the label of the node in the graph

    Returns:
        node: a node in the graph, with specific label. A node is a InstructionNode object.

    """
    gate = copy.deepcopy(input_gate)  # avoid modifying the original gate
    if not isinstance(gate.pos, list):  # if gate.pos is not a list, make it a list
        gate.pos = [gate.pos]

    # use getattr check 'paras' and other attributes if exist. if the attr doesn't exist,return None
    gate.paras = getattr(gate, 'paras', None) or None
    gate.duration = getattr(gate, 'duration', None) or None
    gate.unit = getattr(gate, 'unit', None) or None

    if gate.paras and not isinstance(gate.paras, list):  # if paras is True and not a list, make it a list
        gate.paras = [gate.paras]

    hashable_gate = InstructionNode(gate.name, gate.pos, gate.paras, gate.duration, gate.unit, label=specific_label)
    return hashable_gate


def circuit_to_dag(circuit: QuantumCircuit, measure_flag=True):
    """
    Building a DAG Graph using DAGCircuit from a QuantumCircuit

    Args:
        circuit: a QuantumCircuit object
        measure_flag: whether to add measure_gate node to the dag graph

    Returns:
        g: a DAGCircuit object

    example:
        from circuit_dag_convert import circuit_to_dag, dag_to_circuit, draw_dag
        from quafu import QuantumCircuit

        # Create a quantum circuit as an example that you can modify as needed
        circuit = QuantumCircuit(2)
        circuit.h(0)
        circuit.cnot(0, 1)

        # Build the dag graph
        dag = circuit_to_dag(circuit)  #  dag graph
    """

    # Initialize variables
    i = 0
    qubit_last_use = {}
    g = DAGCircuit()

    # Add the start node
    start_node = -1
    g.add_node(start_node, color="green")

    for gate in circuit.gates:
        # Transform gate to node
        hashable_gate = gate_to_node(gate, specific_label=i)
        i += 1
        g.add_node(hashable_gate, color="blue")

        # Add edges based on qubit_last_use; update last use
        for qubit in hashable_gate.pos:
            prev_node = qubit_last_use.get(qubit, start_node)
            g.add_edge(
                prev_node, hashable_gate, label=f'q{qubit}', color="green" if prev_node == start_node else "black")
            qubit_last_use[qubit] = hashable_gate

    if measure_flag:
        # Add measure_gate node
        measure_pos = copy.deepcopy(circuit.measures)
        measure_gate = InstructionNode("measure", measure_pos, None, None, None, label="m")
        g.add_node(measure_gate, color="blue")

        # Add edges from qubit_last_use to measure_gate
        for qubit in measure_gate.pos:
            prev_node = qubit_last_use.get(qubit, start_node)
            g.add_edge(
                prev_node, measure_gate, label=f'q{qubit}', color="green" if prev_node == start_node else "black")
            qubit_last_use[qubit] = measure_gate

    # Add the end node
    end_node = float('inf')
    g.add_node(end_node, color="red")

    for qubit, last_node in qubit_last_use.items():
        g.add_edge(last_node, end_node, label=f'q{qubit}', color="red")

    # Update DAGCircuit attributes
    g.update_qubits_used()
    g.update_cbits_used()
    g.update_num_instruction_nodes()
    g.update_circuit_qubits(circuit.num)

    return g


def node_to_gate(node_in_dag):
    """
    Transform a gate in a DAG graph to a gate in a circuit which can be added to a QuantumCircuit.

    Args:
        node_in_dag: A node in a DAG graph, an InstructionNode object. node_in_dag.name is uppercase,
                     node_in_dag.pos is a list or a dict. gate_transform supports gates with one or more qubits,
                     excluding measure nodes and special nodes [-1, float('inf'), measure_gate].

    Returns:
        gate: A gate which can be added to a QuantumCircuit in Quafu.

    Example:
        import networkx as nx
        from quafu import QuantumCircuit
        qcircuit = QuantumCircuit(n)

        for gate in nx.topological_sort(dep_graph):
            if gate not in [-1, float('inf')]:
                if gate.name == "measure":
                    qcircuit.measures = gate.pos
                else:
                    qcircuit.gates.append(node_to_gate(gate))
        return qcircuit
    """
    gate_name = node_in_dag.name.lower()
    gate_class = GATE_CLASSES.get(gate_name)

    if not gate_class:
        raise ValueError(f"Gate '{gate_name}' is not supported")

    if gate_name == "barrier":
        return gate_class(node_in_dag.pos)

    # Prepare arguments for gate initialization
    args = node_in_dag.pos
    if node_in_dag.paras:
        args += node_in_dag.paras

    # Handle specific gate types with additional parameters
    if gate_name in ["delay", "xy"]:
        args += [node_in_dag.duration, node_in_dag.unit]

    # Handle multi-qubit gates
    if gate_name in ["mcx", "mcy", "mcz"]:
        control_qubits = node_in_dag.pos[:-1]
        target_qubit = node_in_dag.pos[-1]
        return gate_class(control_qubits, target_qubit)

    return gate_class(*args)


def dag_to_circuit(dep_graph, qubits: int):
    """
    From DAG with Hashable Gates to quafu Gates added to circuit

    Args:
        dep_graph (DAG): DAG with Hashable Gates
        qubits (int): number of qubits

    Returns:
        qcircuit (QuantumCircuit): quafu QuantumCircuit

    Example:
        .. jupyter-execute::

            from circuit_dag import circuit_to_dag, dag_to_circuit, draw_dag
            from quafu import QuantumCircuit

            # Create a quantum circuit as an example that you can modify as needed
            circuit = QuantumCircuit(2)
            circuit.h(0)
            circuit.cnot(0, 1)

            # Build the dag graph
            dep_graph = circuit_to_dag(circuit)  # dag graph

            # Use dag_to_circuit to transform dag graph to a new circuit
            reconstructed_circuit = dag_to_circuit(dep_graph, circuit.num)
    """

    qcircuit = QuantumCircuit(qubits)
    for gate in nx.topological_sort(dep_graph):
        if gate not in [-1, float('inf')]:
            if gate.name == "measure":
                qcircuit.measures = gate.pos
            else:
                qcircuit.gates.append(node_to_gate(gate))
    return qcircuit


def draw_dag(dep_g, output_format="png"):
    """
    Helper function to visualize the DAG

    Args:
        dep_g (DAG): DAG with Hashable Gates
        output_format (str): output format, "png" or "svg"

    Returns:
        img (Image or SVG): show the image of DAG, which is Image(filename="dag.png") or SVG(filename="dag.svg")

    example:
        .. jupyter-execute::
        ex1:
            # directly draw  PNG picture
            draw_dag(dep_g, output_format="png")    # save a png picture "dag.png" and show it in jupyter notebook

            # directly draw  SVG   picture
            draw_dag(dep_g, output_format="svg")    # save a svg picture "dag.svg" and show it in jupyter notebook

        ex2:
            # generate   PNG  picture
            img_png = draw_dag(dep_g, output_format="png")

            # generate   SVG  picture
            img_svg = draw_dag(dep_g, output_format="svg")

            # show PNG picture
            img_png

            # show SVG picture
            img_svg
    """
    from IPython.display import Image, SVG

    try:
        import pygraphviz
        use_pygraphviz = True
    except ImportError:
        use_pygraphviz = False

    if use_pygraphviz:
        # Use pygraphviz
        from networkx.drawing.nx_agraph import write_dot
        write_dot(dep_g, "dag.dot")
        G = pygraphviz.AGraph("dag.dot")
        G.layout(prog="dot")

        if output_format == "png":
            G.draw("dag.png")
            return Image(filename="dag.png")
        elif output_format == "svg":
            G.draw("dag.svg")
            return SVG(filename="dag.svg")
        else:
            raise ValueError("Unsupported output format: choose either 'png' or 'svg'")
    else:
        try:
            # Use graphviz
            from graphviz import Digraph
        except ImportError:
            raise ImportError("Neither 'graphviz' nor 'pygraphviz' are installed.")

        dot = Digraph()

        # Add nodes and edges to the Digraph object based on the DAG
        for node in dep_g.nodes(data=True):
            dot.node(str(node[0]), color=node[1].get("color", "black"), style="solid")

        for edge in dep_g.edges(data=True):
            dot.edge(str(edge[0]), str(edge[1]), color=edge[2].get("color", "black"), label=edge[2].get("label", ""))

        # Render the graph to a file
        if output_format == "png":
            dot.format = 'png'
            dot.render("dag")
            return Image(filename="dag.png")
        elif output_format == "svg":
            dot.format = 'svg'
            dot.render("dag")
            return SVG(filename="dag.svg")
        else:
            raise ValueError("Unsupported output format: choose either 'png' or 'svg'")


def nodelist_to_dag(op_nodes: List[Any]) -> DAGCircuit:
    """
    Transform a list of operation nodes into a DAGCircuit.

    Args:
        op_nodes (List[Any]): A list of operation nodes.

    Returns:
        DAGCircuit: The constructed DAGCircuit object.
    """

    # Initialize variables
    qubit_last_use = {}
    g = DAGCircuit()

    # Add the start node
    start_node = -1
    g.add_node(start_node, color="green")

    for op_node in op_nodes:
        # Transform gate to node
        hashable_gate = copy.deepcopy(op_node)
        g.add_node(hashable_gate, color="blue")

        # Add edges based on qubit_last_use and update last use
        for qubit in hashable_gate.pos:
            prev_node = qubit_last_use.get(qubit, start_node)
            g.add_edge(
                prev_node, hashable_gate, label=f'q{qubit}', color="green" if prev_node == start_node else "black")
            qubit_last_use[qubit] = hashable_gate

    # Add the end node
    end_node = float('inf')
    g.add_node(end_node, color="red")

    for qubit, last_node in qubit_last_use.items():
        g.add_edge(last_node, end_node, label=f'q{qubit}', color="red")

    # Update DAGCircuit attributes
    g.qubits_used = g.update_qubits_used()
    g.cbits_used = g.update_cbits_used()
    g.num_instruction_nodes = g.update_num_instruction_nodes()

    return g


def nodelist_qubit_mapping_dict(nodes_list):
    """
    Generate a mapping dictionary for qubits used by the nodes to new qubits.

    Args:
        nodes_list: a list of nodes

    Returns:
        nodes_qubit_mapping_dict: a dict where keys are the qubits used by the nodes and values are the new qubits
    """
    nodes_list_qubits_used = set()
    for node in nodes_list:
        if hasattr(node, 'pos') and node.pos is not None:
            nodes_list_qubits_used.update(node.pos)

    mapping_pos = list(range(len(nodes_list_qubits_used)))
    # Create the mapping dictionary
    nodes_qubit_mapping_dict = dict(zip(sorted(nodes_list_qubits_used), mapping_pos))

    return nodes_qubit_mapping_dict


def nodelist_qubit_mapping_dict_reverse(nodes_list):
    """
    Generate a reverse mapping dictionary for qubits used by the nodes.

    Args:
        nodes_list: a list of nodes

    Returns:
        nodes_qubit_mapping_dict_reverse: a dict where keys are the new qubits and values are the qubits used by the nodes
    """
    # Generate the initial mapping dictionary
    nodes_qubit_mapping_dict = nodelist_qubit_mapping_dict(nodes_list)

    # Reverse the mapping
    nodes_qubit_mapping_dict_reverse = {value: key for key, value in nodes_qubit_mapping_dict.items()}

    return nodes_qubit_mapping_dict_reverse


def nodes_list_mapping(nodes_list, nodes_qubit_mapping_dict):
    """
    Map qubits in nodes_list to new qubits using the provided mapping dictionary.

    Args:
        nodes_list: The list of instruction nodes.
        nodes_qubit_mapping_dict: The dictionary mapping old qubits to new qubits.

    Returns:
        nodes_list_mapping: The nodes_list after mapping qubits.
    """
    nodes_mapping = []
    for node in nodes_list:
        node_new = copy.deepcopy(node)
        if hasattr(node, 'pos') and node.pos is not None:
            if isinstance(node.pos, list):
                node_new.pos = [nodes_qubit_mapping_dict[qubit] for qubit in node.pos]
            elif isinstance(node.pos, dict):
                node_new.pos = {}
                # The values of the dict are void, so we need to copy the values from the original dict
                for qubit in node.pos:
                    node_new.pos[nodes_qubit_mapping_dict[qubit]] = copy.deepcopy(node.pos[qubit])
        nodes_mapping.append(node_new)
    return nodes_mapping


def copy_dag(dag: DAGCircuit):
    """
    Copy a DAGCircuit object.

    Args:
        dag (DAGCircuit): The DAGCircuit to be copied.

    Returns:
        dag_copy (DAGCircuit): The copied DAGCircuit.
    """
    dag_copy = DAGCircuit()
    dag_copy.circuit_qubits = copy.deepcopy(dag.circuit_qubits)
    dag_copy.qubits_used = copy.deepcopy(dag.qubits_used)
    dag_copy.cbits_used = copy.deepcopy(dag.cbits_used)
    dag_copy.num_instruction_nodes = copy.deepcopy(dag.num_instruction_nodes)
    dag_copy.add_weighted_edges_from(copy.deepcopy(dag.edges(data=True)))
    return dag_copy


def show_dag(dag: DAGCircuit) -> None:
    """
    Show the DAG of a circuit in a figure with proper size in interactive mode.
    This function is used for debugging such as in PyCharm and other IDEs.

    Args:
        dag (DAGCircuit): The DAGCircuit to be displayed.
        dpi (int): Dots per inch, to specify the resolution of the image. Default is 200.
    """
    import matplotlib.pyplot as plt
    from PIL import Image

    draw_dag(dag)
    im = Image.open('dag.png')

    # Convert to array and plot
    data = np.array(im)
    original_width, original_height = data.shape[1], data.shape[0]

    # Display the image with appropriate size and resolution
    dpi = 100
    plt.figure(figsize=(original_width / dpi, original_height / dpi), dpi=dpi)
    plt.imshow(data)
    plt.axis('off')
    plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
    plt.show()



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from quafu import QuantumCircuit

from qsteed.dag.circuit_dag_convert import dag_to_circuit
from qsteed.dag.dagcircuit import DAGCircuit
from qsteed.passes.datadict import DataDict
from qsteed.passflow.passflow import PassFlow
from qsteed.passflow.preset_passflow import PresetPassflow


class Transpiler:
    def __init__(self, passflow: PassFlow = None, initial_model=None):
        self.passflow = passflow
        if initial_model is None:
            self.model = DataDict()
        else:
            self.model = initial_model
            if self.model.datadict is None:
                self.model.datadict = DataDict()

    def set_model(self, new_model):
        self.model = new_model

    def transpile(self, circuit: Union[QuantumCircuit, DAGCircuit], optimization_level: int = 0):
        # give the parameters of the original circuit only once,be careful!
        self.model.datadict['variables'] = circuit.variables

        if self.passflow is None:
            if optimization_level in [0, 1, 2, 3]:
                self.passflow = PresetPassflow(self.model._backend.get_property('basis_gates'),
                                               optimization_level=optimization_level).get_passflow()
            else:
                raise ValueError("Error: The value of optimization_level is between [0,3].")

        for pass_instance in self.passflow.passes:
            if hasattr(pass_instance, 'set_model'):
                pass_instance.set_model(self.model)

            circuit = pass_instance.run(circuit)

            if hasattr(pass_instance, 'get_model'):
                self.model = pass_instance.get_model()

        if isinstance(circuit, DAGCircuit):
            circuit = dag_to_circuit(circuit, circuit.circuit_qubits)

        return circuit



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from qsteed.parallelmanager.parallel_circuits import parallel_process_circuits
from qsteed.transpiler.transpiler import Transpiler
from tests.shared_utils import get_passflow, get_initial_model, get_random_circuit


def create_random_circuits(num_circuits, gates_number: int = 20, depth: int = None):
    circuits = []
    for _ in range(num_circuits):
        qc = get_random_circuit(gates_number=gates_number, depth=depth)
        circuits.append(qc)
    return circuits


class TestParallelTranspile:

    def test_parallel_transpile(self):
        """Create a list of random circuits and transpile them in parallel and serial respectively."""
        num_circuits = 4
        gates_number = 20  # Only one parameter, either the gates_number or the depth, needs to be specified,
        depth = 100  # with preference given to the gates_number if both are provided.
        num_processes = None  # Number of processes
        circuits = create_random_circuits(num_circuits, gates_number=gates_number, depth=depth)
        passflow = get_passflow()
        initial_model = get_initial_model()

        # Parallel transpilation
        t = time.time()
        transpiled_circuits = parallel_process_circuits(circuits, passflow, initial_model, num_processes)
        parallel_time = time.time() - t
        assert len(transpiled_circuits) == num_circuits
        for tc in transpiled_circuits:
            assert tc is not None
        print("Parallel transpilation time:", parallel_time)

        # Serial transpilation
        t = time.time()
        for circuit in circuits:
            transpiler = Transpiler(passflow, initial_model)
            transpiled_circuit = transpiler.transpile(circuit)
            assert transpiled_circuit is not None
        serial_time = time.time() - t
        print("Serial transpilation time:", serial_time)

        # Calculate the ratio of serial transpilation time to parallel transpilation time
        print("Serial transpilation time/parallel transpilation time:", serial_time / parallel_time)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from qsteed.passes.decomposition.utils import matrix_utils as mu


def thinCSD(q1, q2):
    p = q1.shape[0]
    # print("the size of q1/q2: {}".format(p))

    u1, c, v1 = np.linalg.svd(q1)
    v1d = np.conjugate(v1).T
    c = np.flip(c)

    cm = np.zeros((p, p), dtype=complex)
    np.fill_diagonal(cm, c)

    u1 = np.fliplr(u1)
    v1d = np.fliplr(v1d)

    q2 = q2 @ v1d

    # find the biggest index of c[k] <= 1/np.sqrt(2)
    k = 0
    for i in range(1, p):
        if c[i] <= 1 / np.sqrt(2):
            k = i

    k = k + 1
    # print("the k size: {}".format(k))

    u2, _ = np.linalg.qr(q2[:, 0:k], mode='complete')
    # u2, _= np.linalg.qr(q2, mode='complete')
    # print("the size of u2: {}".format(u2.shape))
    # print("the u2 matrix: {}".format(u2))
    s = np.conjugate(u2).T @ q2
    # print("the size of s: {}".format(s.shape))
    # print("the s matrix: {}".format(np.real(s)))

    if k < p:
        r2 = s[k:p, k:p]
        # print("the size of rs: {}".format(r2.shape))
        ut, ss, vt = np.linalg.svd(r2)
        vtd = np.conjugate(vt).T
        s[k:p, k:p] = np.diag(ss)
        cm[:, k:p] = cm[:, k:p] @ vtd
        u2[:, k:p] = u2[:, k:p] @ ut
        v1d[:, k:p] = v1d[:, k:p] @ vtd

        w = cm[k:p, k:p]
        z, r = np.linalg.qr(w, mode='complete')
        cm[k:p, k:p] = r
        u1[:, k:p] = u1[:, k:p] @ z

    for i in range(p):
        if np.real(cm[i, i]) < 0:
            cm[i, i] = -cm[i, i]
            u1[:, i] = -u1[:, i]
        if np.real(s[i, i]) < 0:
            s[i, i] = -s[i, i]
            u2[:, i] = -u2[:, i]

    return u1, u2, v1d, cm, s


def fatCSD(matrix):
    """
    U = [q1, U01] = [u1    ][c  s][v1  ]
        [q2, U11] = [    u2][-s c][   v2]
    """
    # print(matrix)
    U00, U01, U10, U11 = mu.split_matrix(matrix)

    L0, L1, R0, cc, ss = thinCSD(U00, U10)
    R0 = np.conjugate(R0).T
    ss = -ss

    # get the v2
    R1 = np.zeros_like(R0)
    p = R1.shape[0]
    for j in range(p):
        if np.abs(ss[j, j]) > np.abs(cc[j, j]):
            L0d = np.conjugate(L0).T
            tmp = L0d @ U01
            R1[j, :] = tmp[j, :] / ss[j, j]
        else:
            L1d = np.conjugate(L1).T
            tmp = L1d @ U11
            R1[j, :] = tmp[j, :] / cc[j, j]

    assert mu.is_approx(L0 @ cc @ R0, U00)
    assert mu.is_approx(-L1 @ ss @ R0, U10)
    assert mu.is_approx(L0 @ ss @ R1, U01)
    assert mu.is_approx(L1 @ cc @ R1, U11)

    zeros_m = np.zeros_like(L0)
    L = mu.stack_matrices(L0, zeros_m, zeros_m, L1)
    D = mu.stack_matrices(cc, ss, -ss, cc)
    R = mu.stack_matrices(R0, zeros_m, zeros_m, R1)
    assert mu.is_approx(matrix, L @ D @ R)

    return L0, L1, R0, R1, cc, ss  # L0, L1 is unitary



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "dill>=0.3.8",
    "ipython>=8.14.0",
    "matplotlib>=3.5.2",
    "numpy>=1.20.3,<2.0.0",
    "networkx>=2.6.3",
    "scipy>=1.8.1",
    "Pillow>=10.4.0",
    "rich>=13.7.1",
    "graphviz>=0.14.2",
    "tabulate>=0.9.0",
]

setup(
    name="qsteed",
    version="0.1.2",
    author="Xuhongze",
    author_email="xhzby1995@163.com",
    url="https://github.com/BAQIS-Quantum/QSteed",
    description="Quantum Software of Compilation for supporting Real Quantum device",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["qsteed", "quantum compiling", "quantum computing", "quantum operating system"],
    install_requires=requirements,
    packages=find_packages(),
    include_package_data=True,
    extras_require={"tests": ["pytest"]},
    python_requires=">=3.10",
    license="Apache-2.0 License",
)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from quafu import QuantumCircuit

from qsteed.passes.decomposition.ZYZ_decompose import zyz_decomposition
from qsteed.passes.decomposition.utils.matrix_utils import general_kron, general_CNOT


def gates_list_to_circuit(gates_list, nqubit):
    """ Convert gates_list to circuit.

    Args:
        gates_list (list): UnitaryDecompose.gates_list in decomposition.py
        nqubit (int): number of qubits

    Returns:
        circuit: pyquafu QuantumCircuit class
    """
    circuit = QuantumCircuit(nqubit)
    for g in gates_list:
        if g[2] == 'CX':
            circuit.cnot(g[1][0], g[1][1])
        elif g[2] == 'RY':
            circuit.ry(g[1][0], g[3].real)
        elif g[2] == 'RZ':
            circuit.rz(g[1][0], g[3].real)
        elif g[2] == 'U':
            gamma, beta, alpha, global_phase = zyz_decomposition(g[0])
            circuit.rz(g[1][0], gamma)
            circuit.ry(g[1][0], beta)
            circuit.rz(g[1][0], alpha)
        else:
            pass
    return circuit


def gates_list_to_unitary(num_qubit, gates_list):
    """ Convert gates_list to unitary.

    Args:
        num_qubit (int): qubits number of circuit
        gates_list (list): UnitaryDecompose.gates_list in decomposition.py

    Returns:
        unitary: np.array
    """
    unitary_list = []
    for g in gates_list:
        operator = g[0]
        inds = g[1]
        # print(operator, len(operator[::]))
        # assert len(operator) == 2 ** len(inds)

        if len(inds) == 1:
            Ag = general_kron(operator, inds[0], num_qubit)
        elif len(inds) == 2:
            Ag = general_CNOT(num_qubit, inds[0], inds[1])
        else:
            raise (ValueError("The gate must be one/two qubit gate"))
        unitary_list.append(Ag)

    nsize = 2 ** num_qubit
    unitary = np.eye(nsize)
    for u in unitary_list[::-1]:
        unitary = unitary @ u

    return unitary


def circuit_to_unitary(circuit):
    """ Convert gates_list to unitary.

    Args:
        circuit (QuantumCircuit): pyquafu QuantumCircuit class

    Returns:
        unitary: np.array
    """
    num_qubit = circuit.num
    layered_gates = circuit.layered_circuit().T[1:]
    unitary_list = []
    # reverse circuit
    for gates in layered_gates[::-1]:
        for g in gates:
            if g:
                if isinstance(g.pos, int):
                    unit = general_kron(g.matrix, g.pos, num_qubit)
                elif isinstance(g.pos, list) and len(g.pos) == 1:
                    unit = general_kron(g.matrix, g.pos[0], num_qubit)
                elif isinstance(g.pos, list) and g.name == "CX":
                    unit = general_CNOT(num_qubit, g.pos[0], g.pos[1])
                else:
                    raise (ValueError("The gate must be one-qubit gate or CNOT gate."))
                unitary_list.append(unit)

    matrix_size = 2 ** num_qubit
    unitary = np.eye(matrix_size)
    for u in unitary_list:
        unitary = unitary @ u

    return unitary


def get_circuit_depth(num_qubit, gates_list):
    cnot_count = 0
    gate_count_in_each_qubit = np.zeros(num_qubit)
    for g in gates_list:
        operator, inds = g[0], g[1]

        if len(inds) == 2:
            cnot_count += 1

        for ind in inds:
            gate_count_in_each_qubit[ind] += 1

    assert len(gates_list) == (int(sum(gate_count_in_each_qubit)) - cnot_count)
    circuit_depth = int(max(gate_count_in_each_qubit))
    return cnot_count, circuit_depth



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from qsteed.backends.backend import Backend
from qsteed.dag.circuit_dag_convert import circuit_to_dag, dag_to_circuit
from qsteed.parallelmanager.parallel_circuits import parallel_process_circuits
from qsteed.passes.ParameterTuning.parametersubstitution import ParaSubstitution
from qsteed.passes.decomposition.unitary_decompose import UnitaryDecompose
from qsteed.passes.mapping.layout.sabre_layout import SabreLayout
from qsteed.passes.model import Model
from qsteed.passes.optimization.optimization_combine import GateCombineOptimization
from qsteed.passes.optimization.one_qubit_optimization import OneQubitGateOptimization
from qsteed.passes.unroll.unroll_to_2qubit import UnrollTo2Qubit
from qsteed.passes.unroll.unroll_to_basis import UnrollToBasis
from qsteed.transpiler.transpiler import Transpiler
from qsteed.transpiler.transpiler_visualization import TranspilerVis, dynamic_draw
from qsteed.utils.random_circuit import RandomCircuit

__all__ = [
    "Backend",
    "circuit_to_dag",
    "dag_to_circuit",
    "parallel_process_circuits",
    "ParaSubstitution",
    "UnitaryDecompose",
    "SabreLayout",
    "Model",
    "GateCombineOptimization",
    "UnrollTo2Qubit",
    "UnrollToBasis",
    "Transpiler",
    "TranspilerVis",
    "dynamic_draw",
    "RandomCircuit",
    "OneQubitGateOptimization",
]



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx
import numpy as np


def wl_subtree_kernel(g1: nx.Graph, g2: nx.Graph, iteration: int = 3):
    """Compute the Weisfeiler-Lehman Subtree Kernel between two graphs.

    References:
        Shervashidze, N., Schweitzer, P., Leeuwen, E.J.v., Mehlhorn,
        K., Borgwardt, K.M., 2011. Weisfeiler-lehman graph kernels.
        Journal of Machine Learning Research 12, 2539–2561.

    Args:
        g1 (nx.Graph):
        g2 (nx.Graph):
        iteration (int): Maximum height for the subtree kernel computation.

    Returns:
        kernel_value: Subtree kernel similarity between g1 and g2.
    """
    kernel_value = 0
    for i in range(iteration):
        kernel_value = kernel_value + _subtree_kernel(g1, g2)
        g1, g2 = _iteration_graph(g1, g2)
    return kernel_value


def _subtree_kernel(g1: nx.Graph, g2: nx.Graph):
    """Compute the Subtree Kernel between two graphs.
    """
    value = 0
    for items1 in g1.edges.data():
        from1 = g1.nodes[items1[0]]['weight']
        to1 = g1.nodes[items1[1]]['weight']
        if from1 > to1:
            mid = from1
            from1 = to1
            to1 = mid
        weight1 = items1[2]['weight']
        for items2 in g2.edges.data():
            from2 = g2.nodes[items2[0]]['weight']
            to2 = g2.nodes[items2[1]]['weight']
            if from2 > to2:
                mid = from2
                from2 = to2
                to2 = mid
            weight2 = items2[2]['weight']
            if from1 == from2 and to1 == to2:
                value = value + 1 / np.exp((weight1 - weight2) ** 2)
    return value


def _iteration_graph(g1: nx.Graph, g2: nx.Graph):
    """Iteratively generate subtree graph.
    """
    num = 0
    dic = {}
    res1 = {}
    res2 = {}
    for items1 in g1.nodes.data():
        weight1 = items1[1]['weight']
        for items2 in g2.nodes.data():
            weight2 = items2[1]['weight']
            if weight1 > num:
                num = weight1
            if weight2 > num:
                num = weight2
    num = num + 1
    for items1 in g1.nodes.data():
        key = str(items1[1]['weight']) + ","
        rellist = []
        neilist = [n for n in g1.neighbors(items1[0])]
        for i in neilist:
            rellist.append(g1.nodes[i]['weight'])
        rellist.sort()
        for i in rellist:
            key = key + str(i)
        if key not in dic.keys():
            dic[key] = num
            res1[items1[0]] = num
            num = num + 1
        else:
            res1[items1[0]] = dic[key]
    for items2 in g2.nodes.data():
        key = str(items2[1]['weight']) + ","
        rellist = []
        neilist = [n for n in g2.neighbors(items2[0])]
        for i in neilist:
            rellist.append(g2.nodes[i]['weight'])
        rellist.sort()
        for i in rellist:
            key = key + str(i)
        if key not in dic.keys():
            dic[key] = num
            res2[items2[0]] = num
            num = num + 1
        else:
            res2[items2[0]] = dic[key]
    for keys in res1.keys():
        g1.add_node(keys, weight=res1[keys])
    for keys in res2.keys():
        g2.add_node(keys, weight=res2[keys])
    return g1, g2


def fast_subtree_kernel(g1, g2, iteration: int = 3):
    """Compute the Fast Subtree Kernel between two graphs.

    References:
        Shervashidze, N., & Borgwardt, K. M. (2009). Fast subtree kernels on graphs.
        In Advances in Neural Information Processing Systems (pp. 1660-1668).

    Args:
        g1 (nx.Graph):
        g2 (nx.Graph):
        iteration (int): Maximum height for the subtree kernel computation.

    Returns:
        Subtree kernel similarity between g1 and g2.
    """
    labels1 = _get_labels(g1)
    labels2 = _get_labels(g2)
    node_pairs1 = _get_node_pairs(g1)
    node_pairs2 = _get_node_pairs(g2)
    subtrees1 = _get_subtrees(g1, node_pairs1)
    subtrees2 = _get_subtrees(g2, node_pairs2)
    print(labels1, labels2)
    print(node_pairs1, node_pairs2)
    K = np.zeros((len(node_pairs1), len(node_pairs2)))
    for i in range(iteration):
        for j in range(len(node_pairs1)):
            for k in range(len(node_pairs2)):
                if i == 0:
                    if labels1[node_pairs1[j][0]] == labels2[node_pairs2[k][0]]:
                        K[j][k] = 1
                else:
                    subtree1 = subtrees1[node_pairs1[j]]
                    subtree2 = subtrees2[node_pairs2[k]]
                    if nx.is_isomorphic(subtree1, subtree2):
                        K[j][k] += K[j - 1][k - 1]
    return K[-1][-1]


def _get_labels(graph):
    """Get labels of graph.
    """
    labels = []
    for node in graph.nodes():
        label = str(graph.nodes[node]['weight'])
        for neighbor in graph.neighbors(node):
            label += '_' + str(graph.nodes[neighbor]['weight'])
        labels.append(label)
    return labels


def _get_node_pairs(graph):
    """Obtain all node pairs in graph.
    """
    node_pairs = []
    for node in graph.nodes():
        for neighbor in graph.neighbors(node):
            node_pairs.append((node, neighbor))
    return node_pairs


def _get_subtrees(graph, node_pairs):
    """Generate a subgraph based on node pairs.
    """
    subtrees = {}
    for node_pair in node_pairs:
        subtree = nx.Graph()
        subtree.add_edges_from(nx.all_shortest_paths(graph, node_pair[0], node_pair[1]))
        subtrees[node_pair] = subtree
    return subtrees



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from quafu.elements import Instruction
from quafu.elements.element_gates import CYGate, CXGate, SGate, SdgGate

from qsteed.passes.basepass import UnrollPass


class CYToCNOT(UnrollPass):
    """The CYToCNOT pass.

    convert CY gate to {Sdg, S, CNOT}.

    CY gate decomposition rule:
    q[0]  --*--     q[0]  ---------*-------
            |    ≡                 |
    q[1]  --Y--     q[1]  --Sdg----+----S--
    """

    def __init__(self) -> None:
        super().__init__()
        self.original = CYGate.name.lower()
        self.basis = [CXGate.name.lower(), SGate.name.lower(), SdgGate.name.lower()]
        # qc = QuantumCircuit(2)
        # qc.sdg(1)
        # qc.cnot(0,1)
        # qc.s(1)
        # self.circuit = qc

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, CYGate):
            rule.append(SdgGate(op.pos[1]))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(SGate(op.pos[1]))
        else:
            rule.append(op)
        self.rule = rule
        return rule

    # def run(self, circuit: QuantumCircuit) -> QuantumCircuit:
    #     new_circuit = QuantumCircuit(circuit.num)
    #     for op in circuit.gates:
    #         if isinstance(op, CYGate):
    #             new_circuit.add_gate(SdgGate(op.pos[1]))
    #             new_circuit.add_gate(CXGate(op.pos[0], op.pos[1]))
    #             new_circuit.add_gate(SGate(op.pos[1]))
    #         else:
    #             new_circuit.add_gate(op)
    #     return new_circuit



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx
import numpy as np

from qsteed.graph.circuitgraph import draw_graph


class CouplingGraph:

    def __init__(self, coupling_list=None):
        """CouplingGraph initializer

        Args:
            coupling_list (list): [[q0,q1,fidelity], [q1,q0,fidelity], ...]
        """
        self.graph = nx.DiGraph()  # the qubits coupling structure graph of a quantum chip
        self._distance_matrix = None  # adjacency matrix for coupling graph
        self._qubits_list = None  # list of physical qubits in the coupling graph
        self._num_qubits = None  # number of qubits in the coupling graph
        self.is_bidirectional = True  # Determine whether it is bidirectional

        if coupling_list is not None:
            # Sort nodes
            nodeset = sorted(set(item for sublist in coupling_list for item in sublist[0:2]))
            self.graph.add_nodes_from(nodeset)
            if len(coupling_list[0]) == 3:
                for edge in coupling_list:
                    self.graph.add_edges_from([(edge[0], edge[1], {'fidelity': edge[2]})])
            elif len(coupling_list[0]) == 2:
                for edge in coupling_list:
                    self.graph.add_edges_from([(edge[0], edge[1], {'fidelity': 1.0})])
            else:
                raise TypeError("Error: The input coupling_list is wrong.")

        for pair in coupling_list:
            reverse_pair = (pair[1], pair[0])
            if reverse_pair not in coupling_list:
                self.is_bidirectional = False
                break

        self._edge_dict = None
        self._path_fidelity = None

    @property
    def num_qubits(self):
        """Get the number of physical qubits in this coupling graph."""
        if self._num_qubits is None:
            self._num_qubits = len(self.graph)
        return self._num_qubits

    @property
    def edge_dict(self):
        """Get the edge weight dictionary for this graph.

        Returns:
            self._edge_dict (dict): {(qubit1, qubit2): fidelity, ...}
        """
        if self._edge_dict is None:
            self._edge_dict = nx.get_edge_attributes(self.graph, 'fidelity')
        return self._edge_dict

    @property
    def qubits_list(self):
        """Get a list of physical qubits"""
        if self._qubits_list is None:
            self._qubits_list = list(self.graph.nodes())
        return self._qubits_list

    def subgraph(self, node_list):
        """Get the subgraph of this graph.

        Args:
            node_list: node list of subgraph.

        Returns:
            sub_coupling (CouplingGraph)
        """
        sub_coupling = CouplingGraph()
        sub_coupling.graph = self.graph.subgraph(node_list)
        return sub_coupling

    def is_connected(self):
        """Test if the graph is connected.

        Returns:
            nx.is_directed(self.graph) (bool):
            True if the nodes in the graph are connected, False otherwise
        """
        return nx.is_directed(self.graph)

    def neighbors(self, qubit):
        """Returns the nearest neighbors of a given physical qubit in this graph."""
        return self.graph.neighbors(qubit)

    @property
    def distance_matrix(self):
        """Get the distance matrix for the coupling graph."""
        if self._distance_matrix is None:
            if not self.is_connected():
                raise ValueError("Error: This coupling diagram is not connected.")
            self._distance_matrix = nx.floyd_warshall_numpy(self.graph)
        return self._distance_matrix

    def shortest_undirected_path(self, source_qubit, target_qubit):
        """Get the shortest path between source_qubit and target_qubit.

        Args:
            source_qubit: Starting node for path.
            target_qubit: Ending node for path.
        Returns:
            path: Shortest path include both the source and target in the path.
        """
        path = nx.shortest_path(self.graph, source=source_qubit, target=target_qubit)
        if not path:
            raise ValueError("Error: Nodes %s and %s are not connected." % (str(source_qubit), str(target_qubit)))
        return path

    @property
    def path_fidelity(self):
        """Get path fidelity

        The product of the fidelity of all two-qubit gates on the connection path of bits n1 and n2,
        that is, the sum of their respective logarithms.

        Returns: self._path_fidelity
        """
        if self._path_fidelity is None:
            self._path_fidelity = {}
            nodes = len(self.graph.nodes)
            for n1 in range(nodes - 1):
                for n2 in range(n1 + 1, nodes):
                    swap_path = nx.shortest_path(self.graph, n1, n2)
                    if len(swap_path) == 2:  # not need swap
                        self._path_fidelity[(n1, n2)] = np.log(self.edge_dict[(swap_path[0], swap_path[1])])
                    else:
                        fidelity = 0
                        for i in range(len(swap_path) - 2):  # SWAP gates need to be inserted
                            min_f = min(np.log(self.edge_dict[(swap_path[i], swap_path[i + 1])]),
                                        np.log(self.edge_dict[(swap_path[i + 1], swap_path[i])]))
                            max_f = max(np.log(self.edge_dict[(swap_path[i], swap_path[i + 1])]),
                                        np.log(self.edge_dict[(swap_path[i + 1], swap_path[i])]))
                            fidelity += 2 * max_f + min_f
                        fidelity += np.log(self.edge_dict[(swap_path[-1], swap_path[-2])])
                        self._path_fidelity[(n1, n2)] = fidelity

                    swap_path = nx.shortest_path(self.graph, n2, n1)
                    if len(swap_path) == 2:  # not need swap
                        self._path_fidelity[(n2, n1)] = np.log(self.edge_dict[(swap_path[0], swap_path[1])])
                    else:
                        fidelity = 0
                        for i in range(len(swap_path) - 2):  # SWAP gates need to be inserted
                            min_f = min(np.log(self.edge_dict[(swap_path[i], swap_path[i + 1])]),
                                        np.log(self.edge_dict[(swap_path[i + 1], swap_path[i])]))
                            max_f = max(np.log(self.edge_dict[(swap_path[i], swap_path[i + 1])]),
                                        np.log(self.edge_dict[(swap_path[i + 1], swap_path[i])]))
                            fidelity += 2 * max_f + min_f
                            # fidelity += 2 * np.log(self.edge_dict[(swap_path[i], swap_path[i + 1])]) + \
                            #             np.log(self.edge_dict[(swap_path[i + 1], swap_path[i])])
                        fidelity += np.log(self.edge_dict[(swap_path[-1], swap_path[-2])])
                        self._path_fidelity[(n2, n1)] = fidelity

        return self._path_fidelity

    def do_bidirectional(self):
        """Convert unidirectional edges to bidirectional edges."""
        edges_label = self.graph.edges(data=True)
        edges = self.graph.edges()
        for e1, e2, label in edges_label:
            if (e2, e1) not in edges:
                self.graph.add_edges_from([(e2, e1, label)])

        self.is_bidirectional = True

    def draw(self):
        """Draws the coupling map."""
        draw_graph(self.graph)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from quafu.elements.element_gates import CXGate, CZGate, ISwapGate, CPGate

from qsteed.passes.unroll import rules
from qsteed.passes.unroll.rules import *

class_names = rules.__all__

# Use globals() to get class objects and create a dictionary
Rules_dict = {}
# CX_rules = defaultdict(list)
for class_name in class_names:
    class_obj = globals().get(class_name)
    if class_obj is not None:
        if class_obj().original != CXGate.name.lower():
            Rules_dict[class_obj().original] = [class_obj().basis, class_obj()]
        # else:
        #     CX_rules[class_obj().original].append([class_obj().basis, class_obj()])

# Rules_dict = {SwapToCNOT().original: [SwapToCNOT().basis, SwapToCNOT()],
#               RZZToCNOT().original: [RZZToCNOT().basis, RZZToCNOT()]}

# TODO: Conversion of CNOT with other two-qubit gates, such as iswap, cz
CX_rules = {CZGate.name.lower(): {CNOTToCZ().original: [CNOTToCZ().basis, CNOTToCZ()]},
            ISwapGate.name.lower(): {CNOTToISWAP().original: [CNOTToISWAP().basis, CNOTToISWAP()]},
            CPGate(0, 1, 0).name.lower(): {CNOTToCP().original: [CNOTToCP().basis, CNOTToCP()]}}



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import networkx as nx
from networkx.classes.multidigraph import MultiDiGraph

from qsteed.dag.instruction_node import InstructionNode


class DAGCircuit(MultiDiGraph):
    """
    A directed acyclic graph (DAG) representation of a quantum circuit.

    Inherits from MultiDiGraph, which is a directed graph class that can store multiedges.
    Each node in the DAGCircuit represents a quantum operation, and each edge represents a
    dependency between operations. The DAGCircuit is used to optimize the circuit by
    identifying and merging common subcircuits.
    """

    def __init__(self, qubits_used=None, cbits_used=None, incoming_graph_data=None, **attr):
        """
        Create a DAGCircuit.

        Args:
            qubits_used (set[int]): A set of qubits used in the circuit. Defaults to None.
            cbits_used (set[int]): A set of classical bits used in the circuit. Defaults to None.
            circuit_qubits (int): the number of qubits in the circuit. Defaults to None.
            incoming_graph_data: Data to initialize graph. If None (default) an empty graph is created.
            **attr: Keyword arguments passed to the MultiDiGraph constructor.

        Attributes:
            qubits_used (set[int]): A set of qubits used in the circuit.
            cbits_used (set[int]): A set of classical bits used in the circuit.
            num_instruction_nodes (int): The number of instruction nodes in the circuit.
        """
        super().__init__(incoming_graph_data, **attr)

        if qubits_used is None:
            self.qubits_used = set()
        elif isinstance(qubits_used, set):
            self.qubits_used = qubits_used
        else:
            raise ValueError('qubits_used should be a set or None')

        if cbits_used is None:
            self.cbits_used = set()
        elif isinstance(cbits_used, set):
            self.cbits_used = cbits_used
        else:
            raise ValueError('cbits_used should be a set or None')

        self.circuit_qubits = None
        self.num_instruction_nodes = 0

    # Add new methods or override existing methods here.

    def update_circuit_qubits(self, circuit_qubits: int) -> int:
        """
        Update the number of qubits in the quantum circuit.

        Args:
            circuit_qubits (int): The number of qubits in the DAGCircuit.

        Returns:
            int: The updated number of qubits in the DAGCircuit.
        """
        self.circuit_qubits = circuit_qubits
        return self.circuit_qubits

    def update_qubits_used(self) -> set:
        """
        Update and return the set of qubits used in the DAGCircuit.

        The qubits used are determined based on the labels of the edges connected to node -1.
        The qubits are the integer part of the label.

        Returns:
            qubits_used (set): The set of qubits used in the DAGCircuit.
        """
        if -1 not in self.nodes:
            raise ValueError('-1 should be in DAGCircuit, please add it first')

        self.qubits_used = set([int(edge[2]['label'][1:]) for edge in self.out_edges(-1, data=True)])
        return self.qubits_used

    def update_cbits_used(self) -> set:
        """
        Update and return the set of classical bits (cbits) used in the DAGCircuit.

        The cbits used are determined based on the positions of measure nodes in the DAGCircuit.

        Returns:
            cbits_used (set): The set of cbits used in the DAGCircuit.
        """
        for node in self.nodes:
            if hasattr(node, 'name') and node.name == 'measure':
                self.cbits_used = set(node.pos.values())
        return self.cbits_used

    def update_num_instruction_nodes(self) -> int:
        """
        Update and return the number of instruction nodes in the DAGCircuit.

        The number of instruction nodes is calculated as the total number of nodes
        minus 2 (excluding the special nodes -1 and float('inf')).

        Returns:
            num_instruction_nodes (int): The number of instruction nodes in the DAGCircuit.
        """
        if -1 not in self.nodes:
            raise ValueError('-1 should be in DAGCircuit, please add it first')
        if float('inf') not in self.nodes:
            raise ValueError('float("inf") should be in DAGCircuit, please add it first')
        self.num_instruction_nodes = len(self.nodes) - 2

        return self.num_instruction_nodes

    def nodes_labels_resorted(self):
        """
        Update and resort the labels of the nodes in the DAGCircuit.

        This is only for convenience and does not affect the DAGCircuit itself.

        Returns:
            new_dag (DAGCircuit): A new DAGCircuit with nodes' labels resorted.
        """
        # Create a new DAGCircuit
        new_dag = DAGCircuit()

        # Get the sorted list of nodes
        sorted_nodes_list = self.nodes_list()

        # Iterate through the sorted nodes and update labels
        i = 0
        for node in sorted_nodes_list:
            if node.name != 'measure':
                node.label = i
                new_dag.add_instruction_node_end(node)
                i += 1
            else:
                new_dag.add_instruction_node_end(node)

        return new_dag

    def nodes_dict(self) -> dict:
        """
        Return a dictionary of nodes with the node label as key and the node as value,
        excluding nodes with labels -1 and float('inf').

        Returns:
            dict: A dictionary where keys are node labels and values are nodes.
        """
        nodes_dict = {}
        for node in nx.topological_sort(self):
            if node != -1 and node != float('inf'):
                nodes_dict[node.label] = node
        return nodes_dict

    def nodes_list(self) -> list:
        """
        Return a list of nodes without  -1 and float('inf').

        Returns:
            nodes_list (list): A list of nodes excluding -1 and float('inf').
        """
        nodes_list = []
        for node in nx.topological_sort(self):
            if node != -1 and node != float('inf'):
                nodes_list.append(node)
        return nodes_list

    def node_qubits_predecessors(self, node: InstructionNode) -> dict:
        """
        Return a dictionary of qubits to their predecessor nodes for the given node.

        Args:
            node (InstructionNode): A node in the DAGCircuit. The node should not be -1.

        Returns:
            node_qubits_predecessors (dict): A dictionary where keys are qubits and values are predecessor nodes.
        """
        if node not in self.nodes:
            raise ValueError('Node should be in DAGCircuit')
        if node == -1:
            raise ValueError('-1 has no predecessors')

        node_qubits_predecessors = {
            int(edge[2]['label'][1:]): edge[0]
            for edge in self.in_edges(node, data=True)
        }

        return node_qubits_predecessors

    def node_qubits_successors(self, node: InstructionNode) -> dict:
        """
        Return a dictionary of qubits to their successor nodes for the given node.

        Args:
            node (InstructionNode): A node in the DAGCircuit. The node should not be float('inf').

        Returns:
            node_qubits_successors (dict): A dictionary where keys are qubits and values are successor nodes.
        """
        if node not in self.nodes:
            raise ValueError('Node should be in DAGCircuit')
        if node == float('inf'):
            raise ValueError('float("inf") has no successors')

        node_qubits_successors = {
            int(edge[2]['label'][1:]): edge[1]
            for edge in self.out_edges(node, data=True)
        }

        return node_qubits_successors

    def node_qubits_inedges(self, node: InstructionNode) -> dict:
        """
        Return a dictionary of qubits to their incoming edges for the given node.

        Args:
            node (InstructionNode): A node in the DAGCircuit. The node should not be -1.

        Returns:
            dict: A dictionary where keys are qubits and values are incoming edges.
        """
        if node not in self.nodes:
            raise ValueError('Node should be in DAGCircuit')
        if node == -1:
            raise ValueError('-1 has no predecessors')

        inedges = [edge for edge in self.in_edges(node, data=True, keys=True)]
        qubits_labels = [int(edge[2]['label'][1:]) for edge in self.in_edges(node, data=True)]
        node_qubits_inedges = dict(zip(qubits_labels, inedges))
        return node_qubits_inedges

    def node_qubits_outedges(self, node: InstructionNode) -> dict:
        """
        Return a dictionary of qubits to their outgoing edges for the given node.

        Args:
            node (InstructionNode): A node in the DAGCircuit. The node should not be float('inf').

        Returns:
            dict: A dictionary where keys are qubits and values are outgoing edges.
        """
        if node not in self.nodes:
            raise ValueError('Node should be in DAGCircuit')
        if node == float('inf'):
            raise ValueError('float("inf") has no successors')

        outedges = [edge for edge in self.out_edges(node, data=True, keys=True)]
        qubits_labels = [int(edge[2]['label'][1:]) for edge in self.out_edges(node, data=True)]
        node_qubits_outedges = dict(zip(qubits_labels, outedges))
        return node_qubits_outedges

    def remove_instruction_node(self, gate: InstructionNode) -> None:
        """
        Remove a gate from the DAGCircuit and all edges connected to it.
        Add new edges for qubits of the removed gate between all predecessors and successors of the removed gate.

        Args:
            gate (InstructionNode): The gate to be removed from the DAGCircuit. The gate should be in the DAGCircuit,
                                    and should not be -1 or float('inf').

        Raises:
            ValueError: If the gate is not in the DAGCircuit or is -1 or float('inf').
        """
        if gate not in self.nodes:
            raise ValueError('Gate should be in DAGCircuit')
        if gate in {-1, float('inf')}:
            raise ValueError('Gate should not be -1 or float("inf")')

        qubits_predecessors = self.node_qubits_predecessors(gate)
        qubits_successors = self.node_qubits_successors(gate)

        for qubit in gate.pos:
            pred = qubits_predecessors[qubit]
            succ = qubits_successors[qubit]
            if pred != -1 and succ != float('inf'):
                self.add_edge(pred, succ, label=f'q{qubit}')
            elif pred == -1 and succ != float('inf'):
                self.add_edge(pred, succ, label=f'q{qubit}', color='green')
            else:
                self.add_edge(pred, succ, label=f'q{qubit}', color='red')

        self.remove_node(gate)
        self.update_qubits_used()

    def merge_dag(self, other_dag: 'DAGCircuit') -> 'DAGCircuit':
        """
        Merge another DAGCircuit into this DAGCircuit.

        Args:
            other_dag (DAGCircuit): The DAGCircuit to merge into this one.

        Returns:
            DAGCircuit: The merged DAGCircuit.
        """
        # TODO: For two large lines, the error 'networkx.exception.NetworkXUnfeasible:
        #  Graph contains a cycle or graph changed during iteration' is reported.
        #  The labels inside need to be recalibrated to avoid the same nodes.
        if not isinstance(other_dag, DAGCircuit):
            raise ValueError('other_dag should be a DAGCircuit')
        if other_dag is None:
            return self
        if self is None:
            return other_dag

        # For the same qubits (intersection), remove the outgoing edges from the final node of the original DAG
        # and the incoming edges from the initial node of the other DAG,
        # then connect the corresponding tail and head nodes by adding edges
        other_dag_qubits_used = other_dag.update_qubits_used()
        self_qubits_used = self.update_qubits_used()

        # Find intersection of qubits used in both DAGs
        intersect_qubits = self_qubits_used & other_dag_qubits_used
        end_edges_labels_1 = self.node_qubits_inedges(float('inf'))
        start_edges_labels_2 = other_dag.node_qubits_outedges(-1)

        # TODO: For dag with measurement operations, additional processing is required
        if intersect_qubits:
            for qubit in intersect_qubits:
                self.remove_edges_from([end_edges_labels_1[qubit]])
                other_dag.remove_edges_from([start_edges_labels_2[qubit]])
                self.add_edge(end_edges_labels_1[qubit][0], start_edges_labels_2[qubit][1], label=f'q{qubit}')

        # Add nodes and edges from the other DAG to this DAG
        self.add_nodes_from(other_dag.nodes(data=True))
        self.add_edges_from(other_dag.edges(data=True))

        # remove the edges between -1 and float('inf')
        # self.remove_edges_from([edge for edge in self.edges(keys=True) if edge[0] == -1 and edge[1] == float('inf')])

        # Update qubits used in the merged DAG
        self.update_qubits_used()

    def add_instruction_node(self, gate: InstructionNode, predecessors_dict: Dict[int, InstructionNode],
                             successors_dict: Dict[int, InstructionNode]) -> None:
        """
        Add an instruction node into the DAGCircuit, along with all edges connected to it.
        Add new edges for qubits of the new gate between all predecessors and successors of the new gate.

        Args:
            gate (InstructionNode): The instruction node to add. The gate should not be -1 or float('inf').
            predecessors_dict (Dict[int, InstructionNode]): A dictionary of {qubits -> predecessors} of the gate.
            successors_dict (Dict[int, InstructionNode]): A dictionary of {qubits -> successors} of the gate.

        Raises:
            ValueError: If the gate is -1 or float('inf').
        """
        if gate in {-1, float('inf')}:
            raise ValueError('Gate should not be -1 or float("inf")')

        # Remove the edges between the predecessors and successors for the qubits used by the added node
        qubits_pre_out_edges = []
        qubits_suc_in_edges = []
        self.update_qubits_used()

        for qubit in gate.pos:
            if qubit in self.qubits_used:
                pre_out_edges = self.node_qubits_outedges(predecessors_dict[qubit])
                qubits_pre_out_edges.append(pre_out_edges[qubit][:3])  # Use [:3] to get the key of the edge: u, v, k

                suc_in_edges = self.node_qubits_inedges(successors_dict[qubit])
                qubits_suc_in_edges.append(suc_in_edges[qubit][:3])

        qubits_removed_edges = set(qubits_pre_out_edges) | set(qubits_suc_in_edges)
        self.remove_edges_from(qubits_removed_edges)

        # Add the new node and edges
        self.add_node(gate, color="blue")
        for qubit in gate.pos:
            pred = predecessors_dict[qubit]
            succ = successors_dict[qubit]

            if pred == -1:
                self.add_edge(pred, gate, label=f'q{qubit}', color='green')
            else:
                self.add_edge(pred, gate, label=f'q{qubit}')

            if succ == float('inf'):
                self.add_edge(gate, succ, label=f'q{qubit}', color='red')
            else:
                self.add_edge(gate, succ, label=f'q{qubit}')

        # Update qubits
        self.update_qubits_used()

    def add_instruction_node_end(self, gate: InstructionNode):
        """
        Add an instruction node at the end of the DAGCircuit, before the float('inf') node.

        Args:
            gate (InstructionNode): The instruction node to add. The gate should not be -1 or float('inf').

        Raises:
            ValueError: If the gate is -1 or float('inf'), or if the DAGCircuit does not have both -1 and float('inf') nodes.
        """
        if gate in {-1, float('inf')}:
            raise ValueError('Gate should not be -1 or float("inf")')

        if -1 not in self.nodes or float('inf') not in self.nodes:
            # If DAGCircuit is empty, add -1 and float('inf') first
            self.add_nodes_from([(-1, {"color": "green"}), (float('inf'), {"color": "red"})])

            # Add the new node and edges
            self.add_node(gate, color="blue")
            for qubit in gate.pos:
                self.add_edge(-1, gate, label=f'q{qubit}', color='green')
                self.add_edge(gate, float('inf'), label=f'q{qubit}', color='red')

        elif -1 in self.nodes and float('inf') in self.nodes:
            # Get the predecessors_dict of the new node
            inf_predecessors_dict = self.node_qubits_predecessors(float('inf'))
            gate_predecessors_dict = {qubit: inf_predecessors_dict.get(qubit, -1) for qubit in gate.pos}
            gate_successors_dict = {qubit: float('inf') for qubit in gate.pos}

            # Add the new node and edges using add_instruction_node method
            self.add_instruction_node(gate, gate_predecessors_dict, gate_successors_dict)
        else:
            raise ValueError('DAGCircuit should have both -1 and float("inf") nodes at the same time')

    def substitute_node_with_dag(self, gate: InstructionNode, input_dag):
        """
        Substitute a node in the DAGCircuit with another DAGCircuit.

        Args:
            gate (InstructionNode): The node to be substituted.
            input_dag (DAGCircuit): The DAGCircuit to substitute the node with.

        Raises:
            ValueError: If the node is not in the DAGCircuit, is -1 or float('inf'),
                        input_dag is not a DAGCircuit, or the qubits set of input_dag
                        is not the same as the gate's qubits.
        """
        if gate not in self.nodes:
            raise ValueError('node should be in DAGCircuit')
        if gate in {-1, float('inf')}:
            raise ValueError('node should not be -1 or float("inf")')
        if not isinstance(input_dag, DAGCircuit):
            raise ValueError('input_dag should be a DAGCircuit')
        if input_dag is None:
            self.remove_instruction_node(gate)
            return
        if self is None:
            return input_dag

        # Ensure the qubits set of input_dag is the same as the gate‘s qubits
        if input_dag.update_qubits_used() != set(gate.pos):
            raise ValueError('Qubits set of input_dag should be the same as the gate‘s qubits')

        # Find all predecessors and successors of the node to be replaced
        predecessors_dict = self.node_qubits_predecessors(gate)
        successors_dict = self.node_qubits_successors(gate)

        # Remove the node and its edges from the dag graph
        self.remove_node(gate)

        # Add the input_dag into self
        input_dag_startnodes = input_dag.node_qubits_successors(-1)
        input_dag_endnodes = input_dag.node_qubits_predecessors(float('inf'))

        input_dag_qubits = input_dag.update_qubits_used()

        input_dag.remove_node(-1)
        input_dag.remove_node(float('inf'))

        # Add edges between the nodes in input_dag and the predecessors and successors of the original node
        for qubit in input_dag_qubits:
            pred = predecessors_dict[qubit]
            self.add_edge(
                pred, input_dag_startnodes[qubit], label=f'q{qubit}', color='green' if pred == -1 else 'black')
            succ = successors_dict[qubit]
            self.add_edge(
                input_dag_endnodes[qubit], succ, label=f'q{qubit}', color='red' if succ == float('inf') else 'black')

        # Add nodes and edges from input_dag to self
        self.add_nodes_from(input_dag.nodes(data=True))
        self.add_edges_from(input_dag.edges(data=True))

        # Update other attributes of the DAG
        self.update_qubits_used()

    def is_dag(self) -> bool:
        """
        Check if the DAGCircuit is a DAG.

        Returns:
            bool: True if the DAGCircuit is a DAG, False otherwise.
        """
        return nx.is_directed_acyclic_graph(self)

    def get_measure_nodes(self) -> list:
        """
        Get all 'measure' nodes in the DAGCircuit.

        Returns:
            list: A list of 'measure' nodes in the DAGCircuit.
        """
        return [node for node in self.nodes if node not in {-1, float('inf')} and node.name == 'measure']

    def remove_measure_nodes(self, only_last_measure: bool = False) -> None:
        """
        Remove measure nodes in the DAGCircuit.

        Args:
            only_last_measure (bool): If False, remove all measure nodes in the DAGCircuit;
                                      If True, remove the last measure node in the DAGCircuit.
        """
        measure_nodes = self.get_measure_nodes()
        if only_last_measure:
            if measure_nodes:
                self.remove_instruction_node(measure_nodes[-1])
        else:
            for node in measure_nodes:
                self.remove_instruction_node(node)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from quafu.elements import Instruction
from quafu.elements.element_gates import PhaseGate
from quafu.elements.element_gates.rotation import RZGate

from qsteed.passes.basepass import UnrollPass


class PhaseToRZ(UnrollPass):
    """The PhaseToRZ pass.
    convert Phase gate to {RZ}.

    Phase gate decomposition rule:
    q[0] ----P(lambda)---- ≡ U3(0, 0, lambda)
                           ≡ q[0] ----RZ(lambda)---- * global_phase(lambda/2)
    """

    def __init__(self) -> None:
        super().__init__()
        self.parameter_type = 'parameterized_gate'
        self.original = PhaseGate(0, 0).name.lower()
        self.basis = [RZGate(0, 0).name.lower()]
        self.global_phase = 'lambda/2'
        # qc = QuantumCircuit(1)
        # qc.rz(0, lambda)
        # self.circuit = qc

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, PhaseGate):
            if isinstance(op.paras, list):
                paras = op.paras[0]
            else:
                paras = op.paras
            rule.append(RZGate(op.pos[0], paras))
            self.global_phase = paras / 2
        else:
            rule.append(op)
        self.rule = rule
        return rule
